IMPLEMENTATION MODULE DrvDsk;
  (*
    Copyright (C) 1993-1995, Alex Iakovlev & Dimon Maslov
    *********************************************************************
    OpSystems:  MS-DOS 3.1+
    Compilers:  TopSpeed / Stony Brook Modula-2
    ---------------------------------------------------------------------
    D.Maslov    Ver  1.0  Initial version                     February 94
    A.Iakovlev  Ver  1.1  Some minor changes                  February 94
    A.Iakovlev  Ver  2.0  Addresses instead of ARRAY OF BYTE  March 94
    A.Iakovlev  Ver  3.0  Renaming                            March 95
    *********************************************************************
    MS-DOS dependencies:
    ------ functions 3CH, 3DH, 3EH, 3FH, 40H, 41H, 42H, 56H
    ------ call to DOS via interrupt 21H
    TopSpeed compiler dependencies:
    ------ predefined type LONGCARD
    Stony Brook dependencies:
    ------ predefined type LONGCARD
    TopSpeed library dependencies:
    ------ import from SYSTEM: ADDRESS, BYTE, ADR
    Stony Brook library dependencies:
    ------ import from SYSTEM: ADDRESS, BYTE, ADR
  *)

  IMPORT SYSTEM, SrvSys;

  PROCEDURE Dos(VAR reg: SrvSys.Registers): CARDINAL;
    CONST
      DosInterrupt = 21H;
      CarryFlag = 0;
    BEGIN
      IF CarryFlag IN SrvSys.IntRes(reg, DosInterrupt) THEN
        RETURN reg.AX;
      ELSE
        RETURN 0;
      END;
    END Dos;

  PROCEDURE Create(name: ARRAY OF CHAR): File;
    CONST
      CreateCode = 3CH;
      NormalFileAttrs = 20H;
    VAR
      reg: SrvSys.Registers;
    BEGIN
      reg.AH := CreateCode;
      reg.CX := NormalFileAttrs;
      SrvSys.Split(SYSTEM.ADR(name), reg.DS, reg.DX);
      result := Dos(reg);
      IF result = 0 THEN
        RETURN reg.AX;
      ELSE
        RETURN MAX(CARDINAL);
      END;
    END Create;

  PROCEDURE Delete(name: ARRAY OF CHAR);
    CONST
      DeleteCode = 41H;
    VAR
      reg: SrvSys.Registers;
    BEGIN
      reg.AH := DeleteCode;
      SrvSys.Split(SYSTEM.ADR(name), reg.DS, reg.DX);
      result := Dos(reg);
    END Delete;

  PROCEDURE Rename(oldName,newName: ARRAY OF CHAR);
    CONST
      RenameCode = 56H;
    VAR
      reg: SrvSys.Registers;
    BEGIN
      reg.AH := RenameCode;
      SrvSys.Split(SYSTEM.ADR(oldName), reg.DS, reg.DX);
      SrvSys.Split(SYSTEM.ADR(newName), reg.ES, reg.DI);
      result := Dos(reg);
    END Rename;

  PROCEDURE Open(name: ARRAY OF CHAR): File;
    CONST
      OpenCode = 3DH;
      ReadWriteMode = 2H;
    VAR
      reg: SrvSys.Registers;
    BEGIN
      reg.AH := OpenCode;
      reg.AL := ReadWriteMode;
      SrvSys.Split(SYSTEM.ADR(name), reg.DS, reg.DX);
      result := Dos(reg);
      IF result = 0 THEN
        RETURN reg.AX;
      ELSE
        RETURN MAX(CARDINAL);
      END;
    END Open;

  PROCEDURE Close(f: File);
    CONST
      CloseCode = 3EH;
    VAR
      reg: SrvSys.Registers;
    BEGIN
      reg.AH := CloseCode;
      reg.BX := f;
      result := Dos(reg);
    END Close;

  PROCEDURE Exists(name: ARRAY OF CHAR): BOOLEAN;
    BEGIN
      Close(Open(name));
      RETURN result = 0;
    END Exists;

  TYPE SeekType = (FromBegin, FromCurrent, FromEnd);

  PROCEDURE Seek(f: File; t: SeekType; p: LONGCARD): LONGCARD;
    CONST
      SeekCode = 42H;
      Divisor = 10000H;
    VAR
      reg: SrvSys.Registers;
    BEGIN
      reg.AH := SeekCode;
      reg.AL := SYSTEM.BYTE(t);
      reg.BX := f;
      reg.CX := VAL(CARDINAL, p DIV Divisor);
      reg.DX := VAL(CARDINAL, p MOD Divisor);
      result := Dos(reg);
      IF result = 0 THEN
        RETURN VAL(LONGCARD, reg.DX) * Divisor + VAL(LONGCARD, reg.AX);
      ELSE
        RETURN 0;
      END;
    END Seek;

  PROCEDURE Size(f: File): LONGCARD;
    VAR
      currentPosition, fileSize: LONGCARD;
    BEGIN
      (* ----------- set pointer to current position to save it *)
      currentPosition := Seek(f, FromCurrent, 0);
      IF result # 0 THEN RETURN 0 END;
      (* -------------------- set pointer to end of file to obtain size *)
      fileSize := Seek(f, FromEnd, 0);
      IF result # 0 THEN RETURN 0 END;
      (* -------------------------------- restore position & RETURN *)
      IF Seek(f, FromBegin, currentPosition) = currentPosition THEN
        RETURN fileSize;
      ELSE
        RETURN 0;
      END;
    END Size;

  PROCEDURE GetPos(f: File): LONGCARD;
    VAR
      currentPosition: LONGCARD;
    BEGIN
      currentPosition := Seek(f, FromCurrent, 0);
      IF result = 0 THEN
        RETURN currentPosition;
      ELSE
        RETURN 0;
      END;
    END GetPos;

  PROCEDURE SetPos(f: File; p: LONGCARD);
    BEGIN
      p := Seek(f, FromBegin, p);
    END SetPos;

  TYPE FileAccess = (ReadFunc, WriteFunc, TruncFunc);

  PROCEDURE BinIO(op: FileAccess; f: File; count: CARDINAL; buffer: SYSTEM.ADDRESS): CARDINAL;
    CONST
      ReadCode = 3FH;
      WriteCode = 40H;
    VAR
      reg: SrvSys.Registers;
    BEGIN
      IF (op # TruncFunc) AND (count = 0) THEN
        result := 0;
        RETURN count;
      ELSE
        IF op = ReadFunc THEN reg.AH := ReadCode ELSE reg.AH := WriteCode END;
        reg.BX := f;
        reg.CX := count;
        SrvSys.Split(buffer, reg.DS, reg.DX);
        result := Dos(reg);
        IF result = 0 THEN
          RETURN reg.AX;
        ELSE
          RETURN 0;
        END;
      END;
    END BinIO;

  PROCEDURE Truncate(f: File);
    BEGIN
      IF (BinIO(TruncFunc, f, 0, NIL) # 0) AND (result = 0) THEN
        result := MAX(CARDINAL);
      END;
    END Truncate;

  PROCEDURE Write(f: File; src: SYSTEM.ADDRESS; size: CARDINAL): CARDINAL;
    BEGIN
      RETURN BinIO(WriteFunc, f, size, src);
    END Write;

  PROCEDURE Read(f: File; dst: SYSTEM.ADDRESS; size: CARDINAL): CARDINAL;
    BEGIN
      RETURN BinIO(ReadFunc, f, size, dst);
    END Read;

  END DrvDsk.
