IMPLEMENTATION MODULE DrvKbd;
  (*
    Copyright (C) 1993-1995, Alex Iakovlev & Dimon Maslov
    *************************************************************************
    Hardware :  IBM PC, XT, AT or BIOS-compatible
    OpSystems:  MS-DOS & TopSpeed DOS Extender
    -------------------------------------------------------------------------
    A.Iakovlev  Ver  1.0  Initial version                        August 86
    A.Iakovlev  Ver  1.1  Adaptation for LOGITECH ver 1.00       March 87
    A.Iakovlev  Ver  1.2  New InputReady                         October 87
    A.Iakovlev  Ver  1.3  Correction with out pointer            August 88
    A.Iakovlev  Ver  2.0  Procedure Clear                        March 89
    A.Iakovlev  Ver  2.1  Correction in Clear                    March 89
    A.Iakovlev  Ver  3.0  Keys of type CARDINAL                  June 89
    A.Iakovlev  Ver  4.0  Procedure WriteKey                     July 89
    A.Iakovlev  Ver  4.1  BuffSize = 600                         July 89
    A.Iakovlev  Ver  5.0  Background procedure in ReadKey        August 89
    A.Iakovlev  Ver  5.1  DOS system call 7H instead of 8H       August 89
    A.Iakovlev  Ver  5.2  Interface through BIOS                 August 89
    A.Iakovlev  Ver  5.3  Use M2VERS utility                     August 89
    A.Iakovlev  Ver  5.4  Correction in BIOSCheck (JPI)          September 89
    A.Iakovlev  Ver  6.0  Renaming                               January 90
    A.Iakovlev  Ver  6.1  JPI-only version                       January 90
    A.Iakovlev  Ver  7.0  Constants for special keys             June 90
    A.Iakovlev  Ver  8.0  Renaming, no ASCII, arrows processing  July 90
    A.Iakovlev  Ver  8.1  Use switch module OpSystem             July 90
    A.Iakovlev  Ver  9.0  No arrows processing                   February 91
    A.Iakovlev  Ver 10.0  Procedure Get                          February 91
    A.Iakovlev  Ver 11.0  Procedure GetSequence                  February 91
    A.Iakovlev  Ver 12.0  New "space" code                       July 91
    A.Iakovlev  Ver 12.1  Use %T DOS instead of %F OS2           January 92
    A.Iakovlev  Ver 12.2  Version for TopSpeed DOS Extender      January 92
    A.Iakovlev  Ver 13.0  Renaming                               June 93
    A.Iakovlev  Ver 13.1  Import from SYSTEM                     January 94
    A.Iakovlev  Ver 13.2  Use module SrvLow                      January 94
    A.Iakovlev  Ver 13.3  No OS/2 support                        February 94
    A.Iakovlev  Ver 14.0  Procedure KeyHit                       February 94
    A.Iakovlev  Ver 14.1  Use module SrvSys                      February 94
    A.Iakovlev  Ver 15.0  Procedure ScrollLocked                 March 94
    A.Iakovlev  Ver 16.0  Renaming                               March 95
    *************************************************************************
    ROM BIOS & MS-DOS dependencies:
    ------ keyboard driver at INT 16H (functions 0, 1)
    ------ call BIOS via interrupt
    TopSpeed DOS Extender dependencies:
    ------ necessity to initialize the interceptor for the keyboard interrupt
    ------ procedure TSXLIB.InitInt16
  *)

  FROM Cfg IMPORT EXTTS;
  IMPORT SYSTEM, SrvSys, SrvLow;

  (* =========================== Conditional import *)

  (*%T EXTTS*)
  IMPORT TSXLIB;
  (*%E*)

  (* ============================== Hardware/OpSystem Interface *)

  CONST
    KbdInterrupt = 16H;
    ReadFun = 0H;
    CheckFun = 100H;
    ReadFlagsFun = 200H;

  PROCEDURE Peek(): CARDINAL;
    VAR
      reg: SrvSys.Registers;
    BEGIN
      reg.AX := ReadFun;
      SrvSys.Int(reg, KbdInterrupt);
      IF reg.AL = 0 THEN
        RETURN reg.AX;
      ELSE
        RETURN ORD(reg.AL);
      END;
    END Peek;

  PROCEDURE Check(): BOOLEAN;
    CONST
      ZeroFlag = 6;
    VAR
      reg: SrvSys.Registers;
    BEGIN
      reg.AX := CheckFun;
      RETURN NOT (ZeroFlag IN SrvSys.IntRes(reg, KbdInterrupt));
    END Check;

  PROCEDURE Flags(): BITSET;
    VAR
      reg: SrvSys.Registers;
    BEGIN
      reg.AX := ReadFlagsFun;
      SrvSys.Int(reg, KbdInterrupt);
      RETURN BITSET(reg.AX);
    END Flags;

  (* ================================= Public procedures *)

  CONST
    BuffSize = 600;

  VAR
    buff: ARRAY [0 .. BuffSize - 1] OF CARDINAL;  (* ------ cyclic buffer *)
    in, out: CARDINAL;

  PROCEDURE Read(b: PROC; VAR key: CARDINAL);
    BEGIN
      LOOP
        IF in = out THEN
          (* ---------- module buffer is empty *)
          IF Check() THEN
            (* ---------- low-level buffer is not empty *)
            key := Peek();
            buff[in] := key;
            in  := (in + 1) MOD BuffSize;
            out := in;
            EXIT;
          ELSE
            (* ---------- low-level buffer is empty *)
            b();
          END;
        ELSE
          (* ------- module buffer is not empty *)
          key := buff[out];
          out := (out + 1) MOD BuffSize;
          EXIT;
        END;
      END;
    END Read;

  PROCEDURE Get(VAR key: CARDINAL): BOOLEAN;
    BEGIN
      IF in = out THEN
        (* ---------- module buffer is empty *)
        IF Check() THEN
          (* ---------- low-level buffer is not empty *)
          key := Peek();
          buff[in] := key;
          in  := (in + 1) MOD BuffSize;
          out := in;
          RETURN TRUE;
        ELSE
          (* ---------- low-level buffer is empty *)
          RETURN FALSE;
        END;
      ELSE
        (* ------- module buffer is not empty *)
        key := buff[out];
        out := (out + 1) MOD BuffSize;
        RETURN TRUE;
      END;
    END Get;

  PROCEDURE GetSequence(VAR key, count: CARDINAL): BOOLEAN;
    CONST
      MaxCount = 35;
    VAR
      nextKey: CARDINAL;
    BEGIN
      IF Get(key) THEN
        count := 1;
        LOOP
          IF Get(nextKey) THEN
            IF (nextKey = key) AND (count < MaxCount) THEN
              INC(count);
            ELSE
              ReadAgain();
              EXIT;
            END;
          ELSE
            EXIT;
          END;
        END;
        RETURN TRUE;
      ELSE
        RETURN FALSE;
      END;
    END GetSequence;

  PROCEDURE ReadAgain();
    BEGIN
      IF out > 0 THEN
        DEC(out);
      ELSE
        out := BuffSize - 1;
      END;
    END ReadAgain;

  PROCEDURE Write(k: CARDINAL);
    BEGIN
      buff[in] := k;
      in := (in + 1) MOD BuffSize;
    END Write;

  PROCEDURE KeyPressed(): BOOLEAN;
    BEGIN
      RETURN (in # out) OR Check();
    END KeyPressed;

  PROCEDURE ScrollLocked(): BOOLEAN;
    CONST
      Scroll = 4;
    BEGIN
      RETURN Scroll IN Flags();
    END ScrollLocked;

  PROCEDURE KeyHit(): BOOLEAN;
    BEGIN
      RETURN Check();
    END KeyHit;

  PROCEDURE Clear();
    BEGIN
      SrvLow.Fill(SYSTEM.ADR(buff), BuffSize * 2, 0);
      out := 0;
      in := 0;
    END Clear;

  (* =================================== Initialization *)

  BEGIN
  (*%T EXTTS*)
    TSXLIB.InitInt16();
  (*%E*)
    Clear();
  END DrvKbd.
