IMPLEMENTATION MODULE ScrInp;
  (*
    Copyright (C) 1993-1995, Alex Iakovlev & Dimon Maslov
    **********************************************************************
    Compilers:  TopSpeed / Stony Brook Modula-2
    ----------------------------------------------------------------------
    D.Maslov    Ver  1.0  Initial version                      January 91
    A.Iakovlev  Ver  2.0  Renaming pattern -> hatch            February 91
    A.Iakovlev  Ver  3.0  Standard action-returning procedure  February 91
    A.Iakovlev  Ver  3.1  New module DrvKbd                    February 91
    A.Iakovlev  Ver  4.0  Changes in additional attributes     March 91
    A.Iakovlev  Ver  5.0  New actions                          April 91
    A.Iakovlev  Ver  5.1  Screen-adjusted steps                August 91
    A.Iakovlev  Ver  6.0  Use module Res                       August 92
    A.Iakovlev  Ver  7.0  Pragma data(var_enum_size)           August 92
    A.Iakovlev  Ver  7.1  Procedure RequestRD                  August 92
    A.Iakovlev  Ver  7.2  Correction in Get: check CHAR range  August 92
    A.Iakovlev  Ver  8.0  Renaming                             June 93
    A.Iakovlev  Ver  8.1  Use type ScrTxOut.CharSet            January 94
    A.Iakovlev  Ver  8.2  Use module Strings                   January 94
    A.Iakovlev  Ver  9.0  Union all the input primitives       March 95
    **********************************************************************
    TopSpeed compiler dependencies:
    ------ pragmas (*#...*)
    TopSpeed library dependencies:
    ------ import from Strings: Length
    Stony Brook library dependencies:
    ------ import from Strings: Length
  *)

  IMPORT
    Strings,
    Geo, Res, DrvKbd, DrvMou, DrvSnd, Scr, ScrOut, ScrCur;

  (* ================================================== Position input *)

  PROCEDURE Marker(pm: Scr.PIXMAP;
                   VAR a: ScrOut.Attrs; VAR p: Geo.Point;
                   wait: PROC; get: GetPosAction): Res.Status;
    VAR
      mc: ScrCur.MARKERCURSOR;
      p0, p1: Geo.Point;
      dx, dy: INTEGER;
      view: Scr.View;
      res: Res.Status;
    BEGIN
      (* ------------- check the pixmap *)
      IF NOT Scr.Exist(pm) THEN
        RETURN Res.invalidPixmap;
      END;
      (* ---------------------- determine the viewport *)
      Scr.Viewport(pm, view);
      (* ------------------------ store the initial position *)
      p0 := p;
      (* ----------------- check and correct the initial position *)
      IF NOT Geo.PtInRect(p, view.port) THEN
        (* --------- it isn't in the echo area: place it at the centre *)
        Geo.RectCentrePt(view.port, p);
      END;
      (* ----------------------- attempt to create and draw marker cursor *)
      IF ScrCur.NewMarker(pm, a, p, mc) = Res.noMemory THEN
        p := p0;
        RETURN Res.noMemory;
      END;
      ScrCur.ShowMarker(mc);
      (* --------------------- main cycle *)
      LOOP
        (* --------------------------- get the action *)
        CASE get(wait, dx, dy) OF
        | enter:
          (* ------------------------------- end of input *)
          res := Res.done;
          EXIT;
        | escape:
          (* --------------------------------------- break input *)
          p := p0;
          res := Res.break;
          EXIT;
        | move:
          (* ---------------------- moving: attempt to set new point *)
          p1 := p;
          INC(p1.ix, dx);
          INC(p1.iy, dy);
          (* -------------------- test the new point relative the viewport *)
          IF Geo.PtInRect(p1, view.port) THEN
            (* ----------- it is within the viewport: correct result *)
            p := p1;
            (* ------------------------------ move cursor *)
            ScrCur.MoveMarker(mc, p1);
          END;
        END;
      END;
      (* ------------------------ erase cursor *)
      ScrCur.DisposeMarker(mc);
      (* --------------- normal exit *)
      RETURN res;
    END Marker;

  (* ============================================== Rectangle input *)

  TYPE Elements = (LeftUp, RightUp, RightDown, LeftDown);

  PROCEDURE RectCorner(r: Geo.Rectangle; elem: Elements; VAR p: Geo.Point);
  (* determine the position p of the corner elem of the rectangle r *)
    BEGIN
      CASE elem OF
      | LeftUp:
        p.ix := r.ixMin;
        p.iy := r.iyMin;
      | RightUp:
        p.ix := r.ixMax;
        p.iy := r.iyMin;
      | RightDown:
        p.ix := r.ixMax;
        p.iy := r.iyMax;
      | LeftDown:
        p.ix := r.ixMin;
        p.iy := r.iyMax;
      END;
    END RectCorner;

  PROCEDURE Rectangle(pm: Scr.PIXMAP;
                      VAR a: ScrOut.Attrs; VAR r: Geo.Rectangle;
                      wait: PROC; get: GetPosAction): Res.Status;
    CONST
      Frac = 4;
      WholeStart = TRUE;
      ElemStart = LeftUp;
    VAR
      rc: ScrCur.RECTCURSOR;
      mc: ScrCur.MARKERCURSOR;
      ma: ScrOut.Attrs;
      r0, r1: Geo.Rectangle;
      p: Geo.Point;
      step, dx, dy: INTEGER;
      whole: BOOLEAN;
      elem: Elements;
      view: Scr.View;
      res: Res.Status;
    BEGIN
      (* ------------------- check the pixmap *)
      IF NOT Scr.Exist(pm) THEN
        RETURN Res.invalidPixmap;
      END;
      (* -------------------------- determine the viewport *)
      Scr.Viewport(pm, view);
      (* ---------------------------------- store the rectangle *)
      r0 := r;
      (* -------------------------- check and correct the initial rectangle *)
      IF NOT Geo.RectNotEmpty(r) OR NOT Geo.RectInRect(r, view.port) THEN
        (* ---- it is bad or isn't in echo area: set fractional rectangle *)
        r := view.port;
        dx := (r.ixMax - r.ixMin + 1) DIV Frac;
        dy := (r.iyMax - r.iyMin + 1) DIV Frac;
        INC(r.ixMin, dx);
        DEC(r.ixMax, dx);
        INC(r.iyMin, dy);
        DEC(r.iyMax, dy);
      END;
      (* ------------------------- attempt to create and draw cursors *)
      IF ScrCur.NewRectangle(pm, a, r, rc) = Res.noMemory THEN
        r := r0;
        RETURN Res.noMemory;
      END;
      ScrCur.ShowRectangle(rc);
      elem := ElemStart;
      RectCorner(r, elem, p);
      ma := a;
      ma.color := a.auxColor;
      IF ScrCur.NewMarker(pm, ma, p, mc) = Res.noMemory THEN
        ScrCur.DisposeRectangle(rc);
        r := r0;
        RETURN Res.noMemory;
      END;
      whole := WholeStart;
      IF NOT whole THEN
        ScrCur.ShowMarker(mc);
      END;
      (* ------------------------- main cycle *)
      LOOP
        (* ---------------------------- read the keyboard *)
        CASE get(wait, dx, dy) OF
        | enter:
          (* ----------- end of input *)
          res := Res.done;
          EXIT;
        | escape:
          (* ---------------- break input *)
          r := r0;
          res := Res.break;
          EXIT;
        | move:
          (* ---------------- moving: process relative coordinates *)
          r1 := r;
          IF whole THEN
            (* -------------- shifting the rectangle as whole *)
            Geo.ShiftRect(r1, dx, dy);
          ELSE
            (* ------------------ moving one corner of the rectangle *)
            CASE elem OF
            | LeftUp:
              INC(r1.ixMin, dx);
              INC(r1.iyMin, dy);
            | RightUp:
              INC(r1.ixMax, dx);
              INC(r1.iyMin, dy);
            | RightDown:
              INC(r1.ixMax, dx);
              INC(r1.iyMax, dy);
            | LeftDown:
              INC(r1.ixMin, dx);
              INC(r1.iyMax, dy);
            END;
          END;
          (* -------------- test the new rectangle relative to the viewport *)
          IF Geo.RectNotEmpty(r1) AND Geo.RectInRect(r1, view.port) THEN
            (* --------------------------- it is is within the viewport *)
            r := r1;
            (* ------------------------ move cursors *)
            ScrCur.MoveRectangle(rc, r);
            RectCorner(r, elem, p);
            ScrCur.MoveMarker(mc, p);
          END;
        | toggleWhole:
          (* -------------------------- toggle the whole flag *)
          IF whole THEN
            ScrCur.ShowMarker(mc);
            whole := FALSE;
          ELSE
            ScrCur.HideMarker(mc);
            whole := TRUE;
          END;
        | nextCorner:
          (* ----------------------- go to the next corner *)
          IF elem < MAX(Elements) THEN
            INC(elem);
          ELSE
            elem := MIN(Elements);
          END;
          RectCorner(r, elem, p);
          ScrCur.MoveMarker(mc, p);
        END;
      END;
      (* ------------------------- erase cursors *)
      ScrCur.DisposeMarker(mc);
      ScrCur.DisposeRectangle(rc);
      (* ----------------------- normal exit *)
      RETURN res;
    END Rectangle;

  (* ====================================================== String input *)

  TYPE CURSOR = ScrCur.CHARCURSOR;

  PROCEDURE NewCursor(pm: Scr.PIXMAP; a: ScrOut.Attrs; r: Geo.Rectangle; VAR c: CURSOR): BOOLEAN;
    VAR
      p: Geo.Point;
    BEGIN
      a.color := a.auxColor;
      p.ix := r.ixMin; p.iy := r.iyMin;
      RETURN (ScrCur.NewChar(pm, a, p, CHR(a.fillStippleNo), c) = Res.done);
    END NewCursor;

  PROCEDURE ShowCursor(c: CURSOR);
    BEGIN
      ScrCur.ShowChar(c);
    END ShowCursor;

  PROCEDURE MoveCursor(c: CURSOR; r: Geo.Rectangle);
    VAR
      p: Geo.Point;
    BEGIN
      p.ix := r.ixMin; p.iy := r.iyMin;
      ScrCur.MoveChar(c, p);
    END MoveCursor;

  PROCEDURE DisposeCursor(VAR c: CURSOR);
    BEGIN
      ScrCur.DisposeChar(c);
    END DisposeCursor;

  PROCEDURE String(pm: Scr.PIXMAP;
                   VAR a: ScrOut.Attrs; VAR p: Geo.Point; VAR s: ARRAY OF CHAR;
                   wait: PROC; get: GetTxtAction): Res.Status;

    CONST
      AllCharacters = ScrOut.CharSet{" " .. "~"};
      Digits = ScrOut.CharSet{"0" .. "9"};
      Latinica = ScrOut.CharSet{"A" .. "Z", "a" .. "z"};
      Cyrillica = ScrOut.CharSet{CHR(80H) .. CHR(0AFH), CHR(0E0H) .. CHR(0F1H)};

    VAR
      sc: ScrCur.STRINGCURSOR;
      c: CURSOR;
      view: Scr.View;
      rExt, rCurs, rCurs1: Geo.Rectangle;
      delimiters, wordCharacters: ScrOut.CharSet;
      p0: Geo.Point;
      len, pos, pos1, wdt, hgt: CARDINAL;
      dx, dy: INTEGER;
      notBlocked: BOOLEAN;
      insMode: InsMode;
      res: Res.Status;
      ch: CHAR;

    PROCEDURE min(a,b: INTEGER): INTEGER;
      BEGIN
        IF a <= b THEN RETURN a ELSE RETURN b END;
      END min;

    PROCEDURE max(a,b: INTEGER): INTEGER;
      BEGIN
        IF a >= b THEN RETURN a ELSE RETURN b END;
      END max;

    PROCEDURE MoveOrRoll(sh: INTEGER);
      VAR
        shChr, shStr: INTEGER;
      BEGIN
        (* --------------------------------- moving is possible and need? *)
        IF notBlocked AND (sh # 0) THEN
          (* -------- yes: compute number of positions to move character cursor *)
          IF sh > 0 THEN
            shChr := min(sh, (view.port.ixMax - rCurs.ixMax) DIV VAL(INTEGER, wdt));
          ELSIF sh < 0 THEN
            shChr := max(sh, (view.port.ixMin - rCurs.ixMin) DIV VAL(INTEGER, wdt));
          END;
          (* ----------------------- move the character cursor (iff need) *)
          IF shChr # 0 THEN
            Geo.ShiftRect(rCurs, shChr * VAL(INTEGER, wdt), 0);
            MoveCursor(c, rCurs);
          END;
          (* ------------------ move the string cursor (iff need) *)
          shStr := sh - shChr;
          IF shStr # 0 THEN
            DEC(p.ix, shStr * VAL(INTEGER, wdt));
            ScrCur.MoveString(sc, p);
          END;
        END;
      END MoveOrRoll;

    BEGIN
      (* -------------------- check the pixmap *)
      IF NOT Scr.Exist(pm) THEN
        RETURN Res.invalidPixmap;
      END;
      (* -------------------------- set character sets *)
      delimiters := AllCharacters - Latinica;
      wordCharacters := Digits + Latinica + Cyrillica;
      (* ------------------------------ determine the viewport *)
      Scr.Viewport(pm, view);
      (* ---------------------------------- store the old position *)
      p0 := p;
      (* ------------------------ get font size and set initial extent *)
      ScrOut.Dims(a.font, wdt, hgt);
      rExt.ixMin := p.ix;
      rExt.ixMax := p.ix + VAL(INTEGER, (Strings.Length(s) + 1) * wdt) - 1;
      rExt.iyMin := p.iy;
      rExt.iyMax := p.iy + VAL(INTEGER, hgt - 1);
      (* -------------------- check extent and correct the initial position *)
      IF NOT Geo.RectInRect(rExt, view.port) THEN
        (* ---- it isn't in the echo area: place it at the upper left corner *)
        Geo.ShiftRect
        (
          rExt,
          view.port.ixMin - rExt.ixMin,
          view.port.iyMin - rExt.iyMin
        );
        (* -------------------------------------- check the extent once more *)
        IF NOT Geo.RectInRect(rExt, view.port) THEN
          (* ---- it again isn't in the echo area: make visible end of string *)
          Geo.ShiftRect
          (
            rExt,
            view.port.ixMax - rExt.ixMax,
            0
          );
        END;
      END;
      Geo.RectBasePt(rExt, p);
      (* ---------------------- attempt to create and draw string cursor *)
      IF ScrCur.NewString(pm, a, p, s, sc) = Res.noMemory THEN
        p := p0;
        RETURN Res.noMemory;
      END;
      ScrCur.ShowString(sc);
      (* ---------------------- attempt to create and draw character cursor *)
      rCurs := rExt;
      rCurs.ixMin := rCurs.ixMax - VAL(INTEGER, wdt) + 1;
      IF NOT NewCursor(pm, a, rCurs, c) THEN
        ScrCur.DisposeString(sc);
        p := p0;
        RETURN Res.noMemory;
      END;
      ShowCursor(c);
      notBlocked := Geo.RectInRect(rCurs, view.port);
      (* ------------------------------------------- main cycle *)
      len := ScrCur.LengthString(sc);
      insMode := StartInsMode;
      pos := len;
      LOOP
        (* --------------------------- get the action *)
        CASE get(wait, ch, dx, dy) OF
        | enter:
          (* ------------------------------ end of the input *)
          ScrCur.String(sc, a, p, s);
          res := Res.done;
          EXIT;
        | escape:
          (* ---------------------------------- break the input *)
          p := p0;
          res := Res.break;
          EXIT;
        | move:
          (* -------- moving: attempt to set the new cursor *)
          rCurs1 := rCurs;
          Geo.ShiftRect(rCurs1, dx, dy);
          (* --------------- test the new cursor relative the viewport *)
          IF Geo.RectInRect(rCurs1, view.port) THEN
            (* ------ it is within the viewport: update the currents *)
            INC(p.ix, dx);
            INC(p.iy, dy);
            rCurs := rCurs1;
            (* -------------------------- move cursors *)
            ScrCur.MoveString(sc, p);
            MoveCursor(c, rCurs);
          END;
        | char:
          (* -------------------------- character processing *)
          IF NOT (len > HIGH(s)) THEN
            (* ----------------------------- modify the string cursor *)
            ScrCur.ModSubString(sc, ch, pos, insMode = insert);
            (* ------------------------------- move the character cursor *)
            IF pos < len THEN
              (* ---------------------------------- inner position *)
              MoveOrRoll(1);
              len := ScrCur.LengthString(sc);
              INC(pos);
            ELSIF ScrCur.LengthString(sc) > len THEN
              (* --------------- position at end of the string *)
              MoveOrRoll(1);
              INC(len);
              INC(pos);
            END;
          END;
        | toggleInsMode:
          (* ----------------------------------------- toggle insert mode *)
          insMode := VAL(InsMode, ORD(MAX(InsMode)) - ORD(insMode));
          DisposeCursor(c);
          IF NOT NewCursor(pm, a, rCurs, c) THEN
            ScrCur.DisposeString(sc);
            p := p0;
            RETURN Res.noMemory;
          END;
          ShowCursor(c);
        | delChar:
          (* ----------------------------------- delete the character *)
          IF (len > 0) AND (pos < len) THEN
            ScrCur.DelSubString(sc, pos, 1);
            DEC(len);
          END;
        | delLeftChar:
          (* -------------------------- delete the character to the left *)
          IF (len > 0) AND (pos > 0) THEN
            ScrCur.DelSubString(sc, pos - 1, 1);
            MoveOrRoll(-1);
            DEC(len);
            DEC(pos);
          END;
        | delRightWord:
          (* --------------------------------------- delete the right word *)
          IF (len > 0) AND (pos < len) THEN
            pos1 := ScrCur.SkipString(sc, pos, wordCharacters, TRUE);
            IF (pos1 = pos) AND (pos1 < len) THEN INC(pos1) END;
            ScrCur.DelSubString(sc, pos, pos1 - pos);
            DEC(len, pos1 - pos);
          END;
        | delRightStr:
          (* ------------------------------- delete the rest of the string *)
          IF (len > 0) AND (pos < len) THEN
            ScrCur.DelSubString(sc, pos, len - pos);
            len := pos;
          END;
        | delStr:
          (* -------------------------------------- delete the whole string *)
          IF len > 0 THEN
            ScrCur.DelSubString(sc, 0, len);
            MoveOrRoll(- VAL(INTEGER, pos));
            len := 0;
            pos := 0;
          END;
        | nextChar:
          (* ------------------------------------- go to the next character *)
          IF pos < len THEN
            MoveOrRoll(1);
            INC(pos);
          END;
        | prevChar:
          (* --------------------------------- go to the previous character *)
          IF pos > 0 THEN
            MoveOrRoll(-1);
            DEC(pos);
          END;
        | nextWord:
          (* ---------------------------------------- go to the next word *)
          IF pos < len THEN
            pos1 :=
              ScrCur.SkipString
              (
                sc,
                ScrCur.SkipString(sc, pos, wordCharacters, TRUE),
                delimiters,
                TRUE
              );
            MoveOrRoll(pos1 - pos);
            pos := pos1;
          END;
        | prevWord:
          (* ---------------------------------- go to the previous word *)
          IF pos > 0 THEN
            pos1 := ScrCur.SkipString(sc, pos, wordCharacters, FALSE);
            IF pos1 < pos - 1 THEN
              INC(pos1);
            ELSE
              pos1 :=
                ScrCur.SkipString
                (
                  sc,
                  ScrCur.SkipString(sc, pos1, delimiters, FALSE),
                  wordCharacters,
                  FALSE
                );
              IF pos1 > 0 THEN INC(pos1) END;
            END;
            MoveOrRoll(pos1 - pos);
            pos := pos1;
          END;
        | home:
          (* ----------------------------- go to the string beginning *)
          MoveOrRoll(- VAL(INTEGER, pos));
          pos := 0;
        | end:
          (* ------------------------------- go to the string end *)
          MoveOrRoll(len - pos);
          pos := len;
        END;
      END;
      (* ------------------------------- erase character cursor *)
      DisposeCursor(c);
      (* ----------------------------- erase string cursor *)
      ScrCur.DisposeString(sc);
      (* --------------- normal exit *)
      RETURN res;
    END String;

  (* ========================== Standard action-returning procedures *)

  CONST
    XDivisor = 80;
    YDivisor = 50;

  VAR
    sx, sxMax, sy, syMax: CARDINAL;
    graphicsMode: Scr.Type;
    wasGet: BOOLEAN;

  PROCEDURE RequestRD();
    BEGIN
      IF NOT wasGet AND (Scr.Mode() # graphicsMode) THEN
        Scr.Dims(sxMax, syMax);
        sxMax := sxMax DIV XDivisor; sx := sxMax;
        syMax := syMax DIV YDivisor; sy := syMax;
        graphicsMode := Scr.Mode();
        wasGet := TRUE;
      END;
    END RequestRD;

  PROCEDURE GetPos(wait: PROC; VAR dx, dy: INTEGER): Action;
    VAR
      key, count: CARDINAL;
      s: DrvMou.ButtonSet;
    BEGIN
      RequestRD();
      LOOP
        IF DrvKbd.GetSequence(key, count) THEN
          (* ------------------------- keyboard actions *)
          CASE key OF
          | DrvKbd.Home:
            dx := - VAL(INTEGER, count * sx);
            dy := - VAL(INTEGER, count * sy);
            RETURN move;
          | DrvKbd.UpArrow, DrvKbd.CtrlE:
            dx :=   0;
            dy := - VAL(INTEGER, count * sy);
            RETURN move;
          | DrvKbd.PageUp:
            dx :=   VAL(INTEGER, count * sx);
            dy := - VAL(INTEGER, count * sy);
            RETURN move;
          | DrvKbd.LeftArrow, DrvKbd.CtrlS:
            dx := - VAL(INTEGER, count * sx);
            dy :=   0;
            RETURN move;
          | DrvKbd.RightArrow, DrvKbd.CtrlD:
            dx :=   VAL(INTEGER, count * sx);
            dy :=   0;
            RETURN move;
          | DrvKbd.End:
            dx := - VAL(INTEGER, count * sx);
            dy :=   VAL(INTEGER, count * sy);
            RETURN move;
          | DrvKbd.DownArrow, DrvKbd.CtrlX:
            dx :=   0;
            dy :=   VAL(INTEGER, count * sy);
            RETURN move;
          | DrvKbd.PageDown:
            dx :=   VAL(INTEGER, count * sx);
            dy :=   VAL(INTEGER, count * sy);
            RETURN move;
          | DrvKbd.Enter:
            RETURN enter;
          | DrvKbd.Escape:
            RETURN escape;
          | DrvKbd.CtrlV:
            sx := sxMax + 1 - sx;
            sy := syMax + 1 - sy;
          | DrvKbd.Tab:
            RETURN nextCorner;
          | DrvKbd.CtrlH:
            RETURN toggleWhole;
          ELSE
            DrvSnd.Error();
          END;
        ELSIF DrvMou.Get(s, dx, dy) THEN
          (* -------------------------------- mouse actions *)
          IF DrvMou.left IN s THEN
            RETURN enter;
          ELSIF DrvMou.right IN s THEN
            RETURN escape;
          ELSIF DrvMou.middle IN s THEN
            DrvSnd.Error();
          ELSE
            RETURN move;
          END;
        ELSE
          (* ------------- no actions: wait *)
          wait();
        END;
      END;
    END GetPos;

  PROCEDURE GetTxt(wait: PROC; VAR ch: CHAR; VAR dx, dy: INTEGER): Action;
    VAR
      key, count: CARDINAL;
      s: DrvMou.ButtonSet;
    BEGIN
      RequestRD();
      LOOP
        IF DrvKbd.GetSequence(key, count) THEN
          (* --------------------------- keyboard actions *)
          CASE key OF
            (* ----------------------------- position moving *)
          | DrvKbd.CtrlE:
            dx :=   0;
            dy := - VAL(INTEGER, count * sy);
            RETURN move;
          | DrvKbd.CtrlS:
            dx := - VAL(INTEGER, count * sx);
            dy :=   0;
            RETURN move;
          | DrvKbd.CtrlD:
            dx :=   VAL(INTEGER, count * sx);
            dy :=   0;
            RETURN move;
          | DrvKbd.CtrlX:
            dx :=   0;
            dy :=   VAL(INTEGER, count * sy);
            RETURN move;
            (* ------------------------ cursor moving *)
          | DrvKbd.LeftArrow:
            RETURN prevChar;
          | DrvKbd.RightArrow:
            RETURN nextChar;
          | DrvKbd.CtrlLeftArrow:
            RETURN prevWord;
          | DrvKbd.CtrlRightArrow:
            RETURN nextWord;
          | DrvKbd.Home:
            RETURN home;
          | DrvKbd.End:
            RETURN end;
            (* ----------------------- exit *)
          | DrvKbd.Enter:
            RETURN enter;
          | DrvKbd.Escape:
            RETURN escape;
            (* ------------------------ toggles *)
          | DrvKbd.Insert:
            RETURN toggleInsMode;
          | DrvKbd.CtrlV:
            sx := sxMax + 1 - sx;
            sy := syMax + 1 - sy;
            (* --------------------------- deletion *)
          | DrvKbd.Delete:
            RETURN delChar;
          | DrvKbd.BackSpace:
            RETURN delLeftChar;
          | DrvKbd.CtrlT:
            RETURN delRightWord;
          | DrvKbd.CtrlK:
            RETURN delRightStr;
          | DrvKbd.CtrlY:
            RETURN delStr;
          | ORD(" ") .. ORD(MAX(CHAR)):
            ch := CHR(key);
            RETURN char;
          ELSE
          END;
        ELSIF DrvMou.Get(s, dx, dy) THEN
          (* -------------------------------- mouse actions *)
          IF DrvMou.left IN s THEN
            RETURN enter;
          ELSIF DrvMou.right IN s THEN
            RETURN escape;
          ELSIF DrvMou.middle IN s THEN
            DrvSnd.Error();
          ELSE
            RETURN move;
          END;
        ELSE
          (* ------------- no actions: wait *)
          wait();
        END;
      END;
    END GetTxt;

  BEGIN
    wasGet := FALSE;
  END ScrInp.
