IMPLEMENTATION MODULE ScrPal;
  (*
    Copyright (C) 1993-1995, Alex Iakovlev & Dimon Maslov
    **************************************************************************
    Hardware:   IBM PC/XT/AT ROM BIOS
    OpSystem:   MS-DOS & TopSpeed DOS Extender
    Compiler:   TopSpeed / Stony Brook Modula-2
    --------------------------------------------------------------------------
    A.Iakovlev  Ver  1.0  Derived from module Scr                 August 92
    A.Iakovlev  Ver  1.1  Some port functions instead of Lib.Intr August 92
    A.Iakovlev  Ver  1.2  Local procedure SetPal                  August 92
    A.Iakovlev  Ver  1.3  Corr. in VGA4-palette initialization    April 93
    A.Iakovlev  Ver  2.0  Renaming                                June 93
    A.Iakovlev  Ver  3.0  Mass palette shades procedures          June 93
    A.Iakovlev  Ver  4.0  Import from SYSTEM                      January 94
    A.Iakovlev  Ver  4.1  No use SHORTCARD type                   February 94
    A.Iakovlev  Ver  4.2  No OS/2 support                         February 94
    A.Iakovlev  Ver  4.3  Stony Brook support                     February 94
    A.Iakovlev  Ver  5.0  No empty procedures                     March 94
    A.Iakovlev  Ver  6.0  High and true-color modes               October 94
    A.Iakovlev  Ver  7.0  Renaming & shades scaling               November 94
    A.Iakovlev  Ver  8.0  Procedures SetDirectPaletteShades       February 95
    A.Iakovlev  Ver  8.1  Correction in PaletteShade              March 95
    **************************************************************************
    Hardware & BIOS & MS-DOS dependencies:
    ------ video driver at INT 10H (function 10H)
    TopSpeed DOS Extender dependencies:
    ------ necessity to initialize the interceptor for the video interrupt
    ------ procedure TSXLIB.InitInt10
    TopSpeed compiler dependencies:
    ------ structured constants
    ------ one-byte sets
    TopSpeed library dependencies:
    ------ import from SYSTEM: BYTE
    Stony Brook compiler dependencies:
    ------ structured constants
    ------ one-byte sets
    Stony Brook library dependencies:
    ------ import from SYSTEM: BYTE
  *)

  FROM Cfg IMPORT CMPTS, CMPSB, EXTTS;
  IMPORT SYSTEM, SrvSys, SrvLow, ScrDir, Scr;

  (* ===================== Conditional import *)

  (*%T EXTTS*)
  IMPORT TSXLIB;
  (*%E*)

  (* ======================================== Global constants *)

  CONST VideoInterrupt = 10H;

  (* ============================== 16-color EGA palette control functions *)

  CONST
    EGA4ColorsPerPixel = 16;
    EGA4Palettes = 1;
    EGA4Shades = 4;
    EGA4HghBit = 7;
    EGA4LowBit = 6;
    (* ----------------------- *)
    SetPaletteRegFun = 1000H;
    SetPaletteFun = 1002H;

  TYPE
    EGA4PaletteBit = (B, G, R, b, g, r);
    EGA4PaletteBitSet = SET OF EGA4PaletteBit;
    EGA4PaletteType = ARRAY [0 .. EGA4ColorsPerPixel] OF EGA4PaletteBitSet;

  VAR
    curEGA4Palette: EGA4PaletteType;

  PROCEDURE ModifyShade(VAR cs: EGA4PaletteBitSet; pc: Scr.PureColor; sh: CARDINAL);
    BEGIN
      CASE pc OF
      | Scr.red:
        IF EGA4LowBit IN BITSET(sh) THEN INCL(cs, r) ELSE EXCL(cs, r) END;
        IF EGA4HghBit IN BITSET(sh) THEN INCL(cs, R) ELSE EXCL(cs, R) END;
      | Scr.green:
        IF EGA4LowBit IN BITSET(sh) THEN INCL(cs, g) ELSE EXCL(cs, g) END;
        IF EGA4HghBit IN BITSET(sh) THEN INCL(cs, G) ELSE EXCL(cs, G) END;
      | Scr.blue:
        IF EGA4LowBit IN BITSET(sh) THEN INCL(cs, b) ELSE EXCL(cs, b) END;
        IF EGA4HghBit IN BITSET(sh) THEN INCL(cs, B) ELSE EXCL(cs, B) END;
      END;
    END ModifyShade;

  PROCEDURE GetShade(cs: EGA4PaletteBitSet; pc: Scr.PureColor): CARDINAL;
    VAR
      result: BITSET;
    BEGIN
      result := {};
      CASE pc OF
      | Scr.red:
        IF r IN cs THEN INCL(result, EGA4LowBit) END;
        IF R IN cs THEN INCL(result, EGA4HghBit) END;
      | Scr.green:
        IF g IN cs THEN INCL(result, EGA4LowBit) END;
        IF G IN cs THEN INCL(result, EGA4HghBit) END;
      | Scr.blue:
        IF b IN cs THEN INCL(result, EGA4LowBit) END;
        IF B IN cs THEN INCL(result, EGA4HghBit) END;
      END;
      RETURN CARDINAL(result);
    END GetShade;

  PROCEDURE PaletteShadesNmbEGA4(pc: Scr.PureColor): CARDINAL;
    BEGIN
      RETURN EGA4Shades;
    END PaletteShadesNmbEGA4;

  PROCEDURE SetPaletteShadeEGA4(pal: CARDINAL; c: Scr.Pixel; pc: Scr.PureColor; sh: CARDINAL);
    VAR
      reg: SrvSys.Registers;
      e: CARDINAL;
    BEGIN
      e := VAL(CARDINAL, c) MOD EGA4ColorsPerPixel;
      ModifyShade(curEGA4Palette[e], pc, sh);
      reg.AX := SetPaletteRegFun;
      reg.BH := SYSTEM.BYTE(curEGA4Palette[e]);
      reg.BL := SYSTEM.BYTE(e);
      SrvSys.Int(reg, VideoInterrupt);
    END SetPaletteShadeEGA4;

  PROCEDURE PaletteShadeEGA4(pal: CARDINAL; c: Scr.Pixel; pc: Scr.PureColor): CARDINAL;
    VAR
      e: CARDINAL;
    BEGIN
      e := VAL(CARDINAL, c) MOD EGA4ColorsPerPixel;
      RETURN GetShade(curEGA4Palette[e], pc);
    END PaletteShadeEGA4;

  PROCEDURE PaletteShadesSizeEGA4(): CARDINAL;
    BEGIN
      RETURN SIZE(EGA4PaletteType);
    END PaletteShadesSizeEGA4;

  PROCEDURE SetPaletteShadesEGA4(pal: CARDINAL; ptr: SYSTEM.ADDRESS);
    VAR
      reg: SrvSys.Registers;
    BEGIN
      SrvLow.FastMoveInc(ptr, SYSTEM.ADR(curEGA4Palette), SIZE(EGA4PaletteType));
      reg.AX := SetPaletteFun;
      SrvSys.Split(SYSTEM.ADR(curEGA4Palette), reg.ES, reg.DX);
      SrvSys.Int(reg, VideoInterrupt);
    END SetPaletteShadesEGA4;

  PROCEDURE SetDirectPaletteShadesEGA4(pal: CARDINAL);
    VAR
      palette: EGA4PaletteType;
      entry: EGA4PaletteBitSet;
      pc: Scr.PureColor;
      e: CARDINAL;
    BEGIN
      FOR e := 0 TO EGA4ColorsPerPixel - 1 DO
        entry := EGA4PaletteBitSet{};
        FOR pc := MIN(Scr.PureColor) TO MAX(Scr.PureColor) DO
          ModifyShade(entry, pc, ScrDir.DirectShade4(VAL(Scr.Pixel, e), pc));
        END;
        palette[e] := entry;
      END;
      palette[EGA4ColorsPerPixel] := EGA4PaletteBitSet{};
      SetPaletteShadesEGA4(pal, SYSTEM.ADR(palette));
    END SetDirectPaletteShadesEGA4;

  PROCEDURE PaletteShadesEGA4(pal: CARDINAL; ptr: SYSTEM.ADDRESS);
    BEGIN
      SrvLow.FastMoveInc(SYSTEM.ADR(curEGA4Palette), ptr, SIZE(EGA4PaletteType));
    END PaletteShadesEGA4;

  PROCEDURE PalettesNmbEGA4(): CARDINAL;
    BEGIN
      RETURN EGA4Palettes;
    END PalettesNmbEGA4;

  PROCEDURE SetPaletteEGA4(pal: CARDINAL);
    BEGIN
    END SetPaletteEGA4;

  PROCEDURE PaletteEGA4(): CARDINAL;
    BEGIN
      RETURN EGA4Palettes - 1;
    END PaletteEGA4;

  (*%T CMPTS*)

  CONST
    DefaultEGA4Palette = EGA4PaletteType
      (
        EGA4PaletteBitSet {},
        EGA4PaletteBitSet {B},
        EGA4PaletteBitSet {G},
        EGA4PaletteBitSet {G, B},
        EGA4PaletteBitSet {R},
        EGA4PaletteBitSet {R, B},
        EGA4PaletteBitSet {R, g},      (* <--- sic!!! *)
        EGA4PaletteBitSet {R, G, B},
        EGA4PaletteBitSet {r, g, b},
        EGA4PaletteBitSet {r, g, b, B},
        EGA4PaletteBitSet {r, g, b, G},
        EGA4PaletteBitSet {r, g, b, G, B},
        EGA4PaletteBitSet {r, g, b, R},
        EGA4PaletteBitSet {r, g, b, R, B},
        EGA4PaletteBitSet {r, g, b, R, G},
        EGA4PaletteBitSet {r, g, b, R, G, B},
        EGA4PaletteBitSet {}
      );

  (*%E*)

  (*%T CMPSB*)

  CONST
    DefaultEGA4Palette:EGA4PaletteType =
      [
        EGA4PaletteBitSet {},
        EGA4PaletteBitSet {B},
        EGA4PaletteBitSet {G},
        EGA4PaletteBitSet {G, B},
        EGA4PaletteBitSet {R},
        EGA4PaletteBitSet {R, B},
        EGA4PaletteBitSet {R, g},      (* <--- sic!!! *)
        EGA4PaletteBitSet {R, G, B},
        EGA4PaletteBitSet {r, g, b},
        EGA4PaletteBitSet {r, g, b, B},
        EGA4PaletteBitSet {r, g, b, G},
        EGA4PaletteBitSet {r, g, b, G, B},
        EGA4PaletteBitSet {r, g, b, R},
        EGA4PaletteBitSet {r, g, b, R, B},
        EGA4PaletteBitSet {r, g, b, R, G},
        EGA4PaletteBitSet {r, g, b, R, G, B},
        EGA4PaletteBitSet {}
      ];

  (*%E*)

  PROCEDURE StoreAndInitPaletteEGA4();
    BEGIN
      (* ---------- EGA can't read palette => use default settings *)
      curEGA4Palette := DefaultEGA4Palette;
    END StoreAndInitPaletteEGA4;

  PROCEDURE RestorePaletteEGA4();
    VAR
      reg: SrvSys.Registers;
    BEGIN
      reg.AX := SetPaletteFun;
      SrvSys.Split(SYSTEM.ADR(DefaultEGA4Palette), reg.ES, reg.DX);
      SrvSys.Int(reg, VideoInterrupt);
    END RestorePaletteEGA4;

  (* ================================ 16-color VGA palette control functions *)

  CONST
    VGA4ColorsPerPixel = EGA4ColorsPerPixel;
    VGA4Palettes = 16;
    VGA4Shades = 64;
    VGA4ShadesDivisor = ScrDir.MaxShades DIV VGA4Shades;
    (* -------------------------------------------------- *)
    GetPaletteFun = 1009H;
    SetColorRegFun = 1010H;
    SetColorBlockFun = 1012H;
    SetColorPageStateFun = 1013H;
    GetColorBlockFun = 1017H;
    GetColorPageStateFun = 101AH;

  TYPE
    VGA4PureColorArray = ARRAY Scr.PureColor OF SYSTEM.BYTE;
    VGA4PaletteEntry =
      RECORD
        CASE : CARDINAL OF
        | 0:
          acolor: VGA4PureColorArray;
        | 1:
          r, g, b: SYSTEM.BYTE;
        END;
      END;
    VGA4PaletteType = ARRAY [0 .. VGA4ColorsPerPixel - 1] OF VGA4PaletteEntry;
    VGA4PaletteArray = ARRAY [0 .. VGA4Palettes - 1] OF VGA4PaletteType;
    VGA4OrdPalette = ARRAY [0 .. VGA4ColorsPerPixel] OF SYSTEM.BYTE;

  (*%T CMPTS*)

  CONST
    StdPartVGA4Palette = VGA4PaletteType
      (
        VGA4PaletteEntry (VGA4PureColorArray( 0,  0,  0)),
        VGA4PaletteEntry (VGA4PureColorArray( 0,  0, 42)),
        VGA4PaletteEntry (VGA4PureColorArray( 0, 42,  0)),
        VGA4PaletteEntry (VGA4PureColorArray( 0, 42, 42)),
        VGA4PaletteEntry (VGA4PureColorArray(42,  0,  0)),
        VGA4PaletteEntry (VGA4PureColorArray(42,  0, 42)),
        VGA4PaletteEntry (VGA4PureColorArray(42, 21,  0)),
        VGA4PaletteEntry (VGA4PureColorArray(42, 42, 42)),
        VGA4PaletteEntry (VGA4PureColorArray(21, 21, 21)),
        VGA4PaletteEntry (VGA4PureColorArray(21, 21, 63)),
        VGA4PaletteEntry (VGA4PureColorArray(21, 63, 21)),
        VGA4PaletteEntry (VGA4PureColorArray(21, 63, 63)),
        VGA4PaletteEntry (VGA4PureColorArray(63, 21, 21)),
        VGA4PaletteEntry (VGA4PureColorArray(63, 21, 63)),
        VGA4PaletteEntry (VGA4PureColorArray(63, 63, 21)),
        VGA4PaletteEntry (VGA4PureColorArray(63, 63, 63))
      );

  (*%E*)

  (*%T CMPSB*)

  CONST
    StdPartVGA4Palette:VGA4PaletteType =
      [
        [1,  0,  0,  0],
        [1,  0,  0, 42],
        [1,  0, 42,  0],
        [1,  0, 42, 42],
        [1, 42,  0,  0],
        [1, 42,  0, 42],
        [1, 42, 21,  0],
        [1, 42, 42, 42],
        [1, 21, 21, 21],
        [1, 21, 21, 63],
        [1, 21, 63, 21],
        [1, 21, 63, 63],
        [1, 63, 21, 21],
        [1, 63, 21, 63],
        [1, 63, 63, 21],
        [1, 63, 63, 63]
      ];

  (*%E*)

  (*%T CMPTS*)
  CONST Ord = VGA4OrdPalette(0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 0);
  (*%E*)

  (*%T CMPSB*)
  CONST Ord:VGA4OrdPalette = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 0];
  (*%E*)

  VAR
    oldVGA4PaletteState, oldVGA4PaletteNo, curVGA4PaletteNo: CARDINAL;
    oldVGA4PaletteArray, curVGA4PaletteArray: VGA4PaletteArray;
    oldEGA4Palette: EGA4PaletteType;

  PROCEDURE PaletteShadesNmbVGA4(pc: Scr.PureColor): CARDINAL;
    BEGIN
      RETURN VGA4Shades;
    END PaletteShadesNmbVGA4;

  PROCEDURE SetPaletteShadeVGA4(pal: CARDINAL; c: Scr.Pixel; pc: Scr.PureColor; sh: CARDINAL);
    VAR
      reg: SrvSys.Registers;
      e: CARDINAL;
    BEGIN
      pal := pal MOD VGA4Palettes;
      e := VAL(CARDINAL, c) MOD VGA4ColorsPerPixel;
      WITH curVGA4PaletteArray[pal, e] DO
        acolor[pc] := SYSTEM.BYTE(sh DIV VGA4ShadesDivisor);
        reg.AX := SetColorRegFun;
        reg.BX := pal * VGA4ColorsPerPixel + e;
        reg.CH := g;
        reg.CL := b;
        reg.DH := r;
      END;
      SrvSys.Int(reg, VideoInterrupt);
    END SetPaletteShadeVGA4;

  PROCEDURE PaletteShadeVGA4(pal: CARDINAL; c: Scr.Pixel; pc: Scr.PureColor): CARDINAL;
    VAR
      e: CARDINAL;
    BEGIN
      e := VAL(CARDINAL, c) MOD VGA4ColorsPerPixel;
      RETURN
        ORD(curVGA4PaletteArray[pal MOD VGA4Palettes, e].acolor[pc]) *
        VGA4ShadesDivisor;
    END PaletteShadeVGA4;

  PROCEDURE PaletteShadesSizeVGA4(): CARDINAL;
    BEGIN
      RETURN SIZE(VGA4PaletteType);
    END PaletteShadesSizeVGA4;

  PROCEDURE SetPaletteShadesVGA4(pal: CARDINAL; ptr: SYSTEM.ADDRESS);
    VAR
      reg: SrvSys.Registers;
    BEGIN
      pal := pal MOD VGA4Palettes;
      SrvLow.FastMoveInc(ptr, SYSTEM.ADR(curVGA4PaletteArray[pal]), SIZE(VGA4PaletteType));
      reg.AX := SetColorBlockFun;
      reg.BX := pal * VGA4ColorsPerPixel;
      reg.CX := VGA4ColorsPerPixel;
      SrvSys.Split(SYSTEM.ADR(curVGA4PaletteArray), reg.ES, reg.DX);
      SrvSys.Int(reg, VideoInterrupt);
    END SetPaletteShadesVGA4;

  PROCEDURE SetDirectPaletteShadesVGA4(pal: CARDINAL);
    VAR
      palette: VGA4PaletteType;
      entry: VGA4PaletteEntry;
      pc: Scr.PureColor;
      e: CARDINAL;
    BEGIN
      FOR e := 0 TO VGA4ColorsPerPixel - 1 DO
        FOR pc := MIN(Scr.PureColor) TO MAX(Scr.PureColor) DO
          entry.acolor[pc] := SYSTEM.BYTE(ScrDir.DirectShade4(VAL(Scr.Pixel, e), pc) DIV VGA4ShadesDivisor);
        END;
        palette[e] := entry;
      END;
      SetPaletteShadesVGA4(pal, SYSTEM.ADR(palette));
    END SetDirectPaletteShadesVGA4;

  PROCEDURE PaletteShadesVGA4(pal: CARDINAL; ptr: SYSTEM.ADDRESS);
    BEGIN
      SrvLow.FastMoveInc(SYSTEM.ADR(curVGA4PaletteArray[pal MOD VGA4Palettes]), ptr, SIZE(VGA4PaletteType));
    END PaletteShadesVGA4;

  PROCEDURE SetPal(pal: CARDINAL);
    VAR
      reg: SrvSys.Registers;
    BEGIN
      curVGA4PaletteNo := pal MOD VGA4Palettes;
      reg.AX := SetColorPageStateFun;
      reg.BH := SYSTEM.BYTE(curVGA4PaletteNo);
      reg.BL := 1;
      SrvSys.Int(reg, VideoInterrupt);
    END SetPal;

  PROCEDURE PalettesNmbVGA4(): CARDINAL;
    BEGIN
      RETURN VGA4Palettes;
    END PalettesNmbVGA4;

  PROCEDURE SetPaletteVGA4(pal: CARDINAL);
    BEGIN
      SetPal(pal);
    END SetPaletteVGA4;

  PROCEDURE PaletteVGA4(): CARDINAL;
    BEGIN
      RETURN curVGA4PaletteNo;
    END PaletteVGA4;

  PROCEDURE StoreAndInitPaletteVGA4();
    CONST
      InitPaletteNo = 0;
    VAR
      reg: SrvSys.Registers;
    BEGIN
      (* ------------------------ store EGA palette *)
      reg.AX := GetPaletteFun;
      SrvSys.Split(SYSTEM.ADR(oldEGA4Palette), reg.ES, reg.DX);
      SrvSys.Int(reg, VideoInterrupt);
      (* -------------------------- store color page state *)
      reg.AX := GetColorPageStateFun;
      SrvSys.Int(reg, VideoInterrupt);
      oldVGA4PaletteState := ORD(reg.BL);
      oldVGA4PaletteNo := ORD(reg.BH);
      (* -------------------------------- set color page state *)
      reg.AX := SetColorPageStateFun;
      reg.BX := 100H;
      SrvSys.Int(reg, VideoInterrupt);
      SetPal(InitPaletteNo);
      (* ---------------------------------- set ordinal EGA palette *)
      reg.AX := SetPaletteFun;
      SrvSys.Split(SYSTEM.ADR(Ord), reg.ES, reg.DX);
      SrvSys.Int(reg, VideoInterrupt);
      (* ------------------------------------------ store VGA palette *)
      reg.AX := GetColorBlockFun;
      reg.BX := 0;
      reg.CX := VGA4Palettes * VGA4ColorsPerPixel;
      SrvSys.Split(SYSTEM.ADR(oldVGA4PaletteArray), reg.ES, reg.DX);
      SrvSys.Int(reg, VideoInterrupt);
      curVGA4PaletteArray := oldVGA4PaletteArray;
      (* ---------------------------- set standard part of VGA palette *)
      curVGA4PaletteArray[0] := StdPartVGA4Palette;
      reg.AX := SetColorBlockFun;
      reg.BX := 0;
      reg.CX := VGA4ColorsPerPixel;
      SrvSys.Split(SYSTEM.ADR(curVGA4PaletteArray), reg.ES, reg.DX);
      SrvSys.Int(reg, VideoInterrupt);
    END StoreAndInitPaletteVGA4;

  PROCEDURE RestorePaletteVGA4();
    VAR
      reg: SrvSys.Registers;
    BEGIN
      (* ------------------- restore VGA palette *)
      reg.AX := SetColorBlockFun;
      reg.BX := 0;
      reg.CX := VGA4Palettes * VGA4ColorsPerPixel;
      SrvSys.Split(SYSTEM.ADR(oldVGA4PaletteArray), reg.ES, reg.DX);
      SrvSys.Int(reg, VideoInterrupt);
      (* ----------------------------- restore color page state *)
      reg.AX := SetColorPageStateFun;
      reg.BH := SYSTEM.BYTE(oldVGA4PaletteState);
      reg.BL := 0;
      SrvSys.Int(reg, VideoInterrupt);
      SetPaletteVGA4(oldVGA4PaletteNo);
      (* --------------------------- restore EGA palette *)
      reg.AX := SetPaletteFun;
      SrvSys.Split(SYSTEM.ADR(oldEGA4Palette), reg.ES, reg.DX);
      SrvSys.Int(reg, VideoInterrupt);
    END RestorePaletteVGA4;

  (* =============================== 256-color VGA palette control functions *)

  CONST
    VGA8ColorsPerPixel = 256;
    VGA8Palettes = 1;
    VGA8Shades = VGA4Shades;
    VGA8ShadesDivisor = VGA4ShadesDivisor;

  TYPE
    VGA8PaletteEntry = VGA4PaletteEntry;
    VGA8PaletteType = ARRAY [0 .. VGA8ColorsPerPixel - 1] OF VGA8PaletteEntry;

  VAR
    oldVGA8Palette, curVGA8Palette: VGA8PaletteType;

  PROCEDURE PaletteShadesNmbVGA8(pc: Scr.PureColor): CARDINAL;
    BEGIN
      RETURN VGA8Shades;
    END PaletteShadesNmbVGA8;

  PROCEDURE SetPaletteShadeVGA8(pal: CARDINAL; c: Scr.Pixel; pc: Scr.PureColor; sh: CARDINAL);
    VAR
      reg: SrvSys.Registers;
      e: CARDINAL;
    BEGIN
      e := VAL(CARDINAL, c) MOD VGA8ColorsPerPixel;
      WITH curVGA8Palette[e] DO
        acolor[pc] := SYSTEM.BYTE(sh DIV VGA8ShadesDivisor);
        reg.AX := SetColorRegFun;
        reg.BX := e;
        reg.CH := g;
        reg.CL := b;
        reg.DH := r;
      END;
      SrvSys.Int(reg, VideoInterrupt);
    END SetPaletteShadeVGA8;

  PROCEDURE PaletteShadeVGA8(pal: CARDINAL; c: Scr.Pixel; pc: Scr.PureColor): CARDINAL;
    VAR
      e: CARDINAL;
    BEGIN
      e := VAL(CARDINAL, c) MOD VGA8ColorsPerPixel;
      RETURN ORD(curVGA8Palette[e].acolor[pc]) * VGA8ShadesDivisor;
    END PaletteShadeVGA8;

  PROCEDURE PaletteShadesSizeVGA8(): CARDINAL;
    BEGIN
      RETURN SIZE(VGA8PaletteType);
    END PaletteShadesSizeVGA8;

  PROCEDURE SetPaletteShadesVGA8(pal: CARDINAL; ptr: SYSTEM.ADDRESS);
    VAR
      reg: SrvSys.Registers;
    BEGIN
      SrvLow.FastMoveInc(ptr, SYSTEM.ADR(curVGA8Palette), SIZE(VGA8PaletteType));
      reg.AX := SetColorBlockFun;
      reg.BX := 0;
      reg.CX := VGA8ColorsPerPixel;
      SrvSys.Split(SYSTEM.ADR(curVGA8Palette), reg.ES, reg.DX);
      SrvSys.Int(reg, VideoInterrupt);
    END SetPaletteShadesVGA8;

  PROCEDURE SetDirectPaletteShadesVGA8(pal: CARDINAL);
    VAR
      palette: VGA8PaletteType;
      entry: VGA8PaletteEntry;
      pc: Scr.PureColor;
      e: CARDINAL;
    BEGIN
      FOR e := 0 TO VGA8ColorsPerPixel - 1 DO
        FOR pc := MIN(Scr.PureColor) TO MAX(Scr.PureColor) DO
          entry.acolor[pc] := SYSTEM.BYTE(ScrDir.DirectShade8(VAL(Scr.Pixel, e), pc) DIV VGA8ShadesDivisor);
        END;
        palette[e] := entry;
      END;
      SetPaletteShadesVGA8(pal, SYSTEM.ADR(palette));
    END SetDirectPaletteShadesVGA8;

  PROCEDURE PaletteShadesVGA8(pal: CARDINAL; ptr: SYSTEM.ADDRESS);
    BEGIN
      SrvLow.FastMoveInc(SYSTEM.ADR(curVGA8Palette), ptr, SIZE(VGA8PaletteType));
    END PaletteShadesVGA8;

  PROCEDURE PalettesNmbVGA8(): CARDINAL;
    BEGIN
      RETURN VGA8Palettes;
    END PalettesNmbVGA8;

  PROCEDURE SetPaletteVGA8(pal: CARDINAL);
    BEGIN
    END SetPaletteVGA8;

  PROCEDURE PaletteVGA8(): CARDINAL;
    BEGIN
      RETURN VGA8Palettes - 1;
    END PaletteVGA8;

  PROCEDURE StoreAndInitPaletteVGA8();
    VAR
      reg: SrvSys.Registers;
    BEGIN
      reg.AX := GetColorBlockFun;
      reg.BX := 0;
      reg.CX := VGA8ColorsPerPixel;
      SrvSys.Split(SYSTEM.ADR(oldVGA8Palette), reg.ES, reg.DX);
      SrvSys.Int(reg, VideoInterrupt);
      curVGA8Palette := oldVGA8Palette;
    END StoreAndInitPaletteVGA8;

  PROCEDURE RestorePaletteVGA8();
    VAR
      reg: SrvSys.Registers;
    BEGIN
      reg.AX := SetColorBlockFun;
      reg.BX := 0;
      reg.CX := VGA8ColorsPerPixel;
      SrvSys.Split(SYSTEM.ADR(oldVGA8Palette), reg.ES, reg.DX);
      SrvSys.Int(reg, VideoInterrupt);
    END RestorePaletteVGA8;

  (* ========================== Empty procedures for high/true-color modes *)

  PROCEDURE PaletteShadesNmbEmpty(pc: Scr.PureColor): CARDINAL;
    BEGIN
      RETURN 1;
    END PaletteShadesNmbEmpty;

  PROCEDURE SetPaletteShadeEmpty(pal: CARDINAL; c: Scr.Pixel; pc: Scr.PureColor; sh: CARDINAL);
    BEGIN
    END SetPaletteShadeEmpty;

  PROCEDURE PaletteShadeEmpty(pal: CARDINAL; c: Scr.Pixel; pc: Scr.PureColor): CARDINAL;
    BEGIN
      RETURN 0;
    END PaletteShadeEmpty;

  PROCEDURE PaletteShadesSizeEmpty(): CARDINAL;
    BEGIN
      RETURN 0;
    END PaletteShadesSizeEmpty;

  PROCEDURE SetPaletteShadesEmpty(pal: CARDINAL; ptr: SYSTEM.ADDRESS);
    BEGIN
    END SetPaletteShadesEmpty;

  PROCEDURE SetDirectPaletteShadesEmpty(pal: CARDINAL);
    BEGIN
    END SetDirectPaletteShadesEmpty;

  PROCEDURE PaletteShadesEmpty(pal: CARDINAL; ptr: SYSTEM.ADDRESS);
    BEGIN
    END PaletteShadesEmpty;

  PROCEDURE PalettesNmbEmpty(): CARDINAL;
    BEGIN
      RETURN 1;
    END PalettesNmbEmpty;

  PROCEDURE SetPaletteEmpty(pal: CARDINAL);
    BEGIN
    END SetPaletteEmpty;

  PROCEDURE PaletteEmpty(): CARDINAL;
    BEGIN
      RETURN 0;
    END PaletteEmpty;

  PROCEDURE StoreAndInitPaletteEmpty();
    BEGIN
    END StoreAndInitPaletteEmpty;

  PROCEDURE RestorePaletteEmpty();
    BEGIN
    END RestorePaletteEmpty;

  (* ====================================== Initialization *)

  BEGIN
    (* ----------- initialize interceptor *)
  (*%T EXTTS*)
    TSXLIB.InitInt10();
  (*%E*)
  END ScrPal.
