IMPLEMENTATION MODULE ScrPl4;
  (*
    Copyright (C) 1993-1995, Alex Iakovlev & Dimon Maslov
    *************************************************************************
    Hardware :  16-color Enhanced Graphics Adapter/ Video Graphics Array
    Compilers:  TopSpeed / Stony Brook Modula-2
    -------------------------------------------------------------------------
    A.Iakovlev  Ver  1.0  Derived from generic module Scr        December 90
    A.Iakovlev  Ver  1.1  Accurate use of byte/word moves        December 90
    A.Iakovlev  Ver  1.2  Use ports via separate modules         May 91
    A.Iakovlev  Ver  1.3  Use VGA write mode 3 in MaskRect       May 91
    A.Iakovlev  Ver  1.4  New module Utils                       May 91
    A.Iakovlev  Ver  1.5  Corrections in MaskRect for w > 1      August 91
    A.Iakovlev  Ver  1.6  Use combined low-level module RM       January 92
    A.Iakovlev  Ver  2.0  Module Res + procs New/Dispose         August 92
    A.Iakovlev  Ver  3.0  InitPage/InitPageAddrs/TermPageAddrs   August 92
    A.Iakovlev  Ver  4.0  Renaming                               June 93
    A.Iakovlev  Ver  5.0  Import from SYSTEM                     January 94
    A.Iakovlev  Ver  5.1  Use SrvSys.Add instead of Lib.IncAddr  January 94
    A.Iakovlev  Ver  5.2  No use SHORTCARD type                  February 94
    A.Iakovlev  Ver  5.3  Stony Brook support                    February 94
    A.Iakovlev  Ver  5.4  No shift operators                     February 94
    A.Iakovlev  Ver  5.5  Proc InitPage: width in bytes          March 94
    A.Iakovlev  Ver  6.0  Color type = Scr.Pixel                 September 94
    A.Iakovlev  Ver  6.1  Use module Mem for heap control        November 94
    *************************************************************************
    Hardware dependencies:
    ------ one bit from four planes per pixel
    ------ pixels positions on the pixmap
    ------ control via ports
    TopSpeed compiler dependencies:
    ------ options (*$...*)
    ------ structured constants
    ------ single-byte sets
    TopSpeed library dependencies:
    ------ import from SYSTEM: BYTE
    Stony Brook compiler dependencies:
    ------ structured constants
    ------ single-byte sets
    Stony Brook library dependencies:
    ------ import from SYSTEM: BYTE
  *)

  FROM Cfg IMPORT CMPTS, CMPSB;
  IMPORT
    SYSTEM,
    Geo, Res, DrvMem, SrvSys, SrvMat, SrvObj, SrvLow,
    ScrTyp, ScrVar, ScrBas, ScrLow, Scr;

  (* ====================================== Type alias names *)

  TYPE
    BS = ScrTyp.BS;
    BP = ScrTyp.BP;

  (* ================================================ Basic constants *)

  CONST
    BitsPerPixel = 1;
    PlanesPerPixel = 4;
    ColorsPerPixel = 16;
    PixelsPerByte = 8 DIV BitsPerPixel;

  (*
    NOTE: Just the same constants are defined in the configuration
    module, but it isn't possible to customize them here.
  *)

  (* ========================================== Screen masks *)

  CONST
    Full = BS{0 .. PixelsPerByte - 1};
    Null = BS{};

  TYPE
    BM = ARRAY [0 .. PixelsPerByte - 1] OF BS;

  (*%T CMPTS*)

  CONST
    LftMask = BM (BS{0 .. 7}, BS{0 .. 6}, BS{0 .. 5}, BS{0 .. 4}, BS{0 .. 3}, BS{0 .. 2}, BS{0 .. 1}, BS{0     });
    RhtMask = BM (BS{     7}, BS{6 .. 7}, BS{5 .. 7}, BS{4 .. 7}, BS{3 .. 7}, BS{2 .. 7}, BS{1 .. 7}, BS{0 .. 7});
    (* -------------------------------------------------------------------- *)
    ShlMask = BM (BS{0}, BS{1}, BS{2}, BS{3}, BS{4}, BS{5}, BS{6}, BS{7});
    ShrMask = BM (BS{7}, BS{6}, BS{5}, BS{4}, BS{3}, BS{2}, BS{1}, BS{0});

  (*%E*)

  (*%T CMPSB*)

  CONST
    LftMask:BM = [BS{0 .. 7}, BS{0 .. 6}, BS{0 .. 5}, BS{0 .. 4}, BS{0 .. 3}, BS{0 .. 2}, BS{0 .. 1}, BS{0     }];
    RhtMask:BM = [BS{     7}, BS{6 .. 7}, BS{5 .. 7}, BS{4 .. 7}, BS{3 .. 7}, BS{2 .. 7}, BS{1 .. 7}, BS{0 .. 7}];
    (* -------------------------------------------------------------------- *)
    ShlMask:BM = [BS{0}, BS{1}, BS{2}, BS{3}, BS{4}, BS{5}, BS{6}, BS{7}];
    ShrMask:BM = [BS{7}, BS{6}, BS{5}, BS{4}, BS{3}, BS{2}, BS{1}, BS{0}];

  (*%E*)

  (* ================================================ Globals *)

  VAR vga: BOOLEAN;

  (* ================================ Line parameters *)

  TYPE
    LineParams =
      RECORD
        byteCount: CARDINAL;
        LM, RM: BS;
      END;

  (*$N,I-,O-*)

  PROCEDURE SetLineParams(rxMin, rxMax: CARDINAL; VAR lp: LineParams);
    BEGIN
      WITH lp DO
        byteCount := rxMax DIV PixelsPerByte - rxMin DIV PixelsPerByte;
        IF byteCount > 0 THEN
          LM := LftMask[rxMin MOD PixelsPerByte];
          RM := RhtMask[rxMax MOD PixelsPerByte];
        ELSE
          LM := LftMask[rxMin MOD PixelsPerByte] * RhtMask[rxMax MOD PixelsPerByte];
          RM := LM;
        END;
      END;
    END SetLineParams;

  (*$F,I=,O=*)

  (* =================================== Full-pixmap functions *)

  PROCEDURE InitPage(w, h: CARDINAL);
    VAR
      bytesPerPage: CARDINAL;
      r: Geo.Rectangle;
    BEGIN
      bytesPerPage := w * h;
      Geo.SetRect(r, 0, (w * PixelsPerByte) - 1, 0, h - 1);
      WITH ScrVar.SPM DO
        number := MAX(CARDINAL) DIV bytesPerPage;
        length := bytesPerPage;
        width := w;
        surf := r;
        view.port := r;
      END;
    END InitPage;

  PROCEDURE InitPageAddrs(sel: CARDINAL);
    VAR
      pl: CARDINAL;
    BEGIN
      FOR pl := 0 TO PlanesPerPixel - 1 DO
        ScrVar.SPM.base[pl].addr := SrvSys.Addr(sel, 0);
      END;
    END InitPageAddrs;

  PROCEDURE TermPageAddrs();
    BEGIN
      ScrVar.SPM.base := ScrTyp.NilBanks;
    END TermPageAddrs;

  PROCEDURE New(r: Geo.Rectangle; VAR pm: PIXMAP): Res.Status;
    VAR
      newPm: PIXMAP;
      longLen: LONGCARD;
      lftByte: INTEGER;
      pl: CARDINAL;
    BEGIN
      IF DrvMem.Alloc(newPm, SIZE(ScrTyp.PixmapDescriptor)) THEN
        WITH newPm^ DO
          (* ------------------- compute width and length of plane *)
          lftByte := SrvMat.Div(r.ixMin, PixelsPerByte);
          width := SrvMat.Div(r.ixMax, PixelsPerByte) - lftByte + 1;
          longLen := VAL(LONGCARD, width) * VAL(LONGCARD, r.iyMax - r.iyMin + 1);
          (* ----------------------------------------- is the plane allocatable? *)
          IF longLen <= VAL(LONGCARD, MAX(CARDINAL)) THEN
            (* --------------------------------- set length of plane *)
            length := VAL(CARDINAL, longLen);
            (* ----------------------- attempt to allocate pixmap planes *)
            base := ScrTyp.NilBanks;
            pl := 0;
            LOOP
              IF pl < PlanesPerPixel THEN
                IF DrvMem.Alloc(base[pl].addr, length) THEN
                  (* --------------------- the plane is allocated *)
                  INC(pl);
                ELSE
                  (* ----------------- fail during plane allocation *)
                  EXIT;
                END;
              ELSE
                (* ----- all planes ok: attempt to allocate buffer *)
                IF DrvMem.Alloc(buffer.addr, width + 1) THEN
                  (* ------------------ success: set other parameters *)
                  number := PlanesPerPixel;
                  WHILE (length > MAX(CARDINAL) DIV number) DO
                    number := number DIV 2;
                  END;
                  surf := r;
                  ix0 := lftByte * PixelsPerByte;
                  iy0 := r.iyMin;
                  view.port := surf;
                  view.clip := TRUE;
                  screen := FALSE;
                  (* ---------- include into the pixmap stack *)
                  next := ScrVar.topPm;
                  ScrVar.topPm := newPm;
                  (* --------------- normal exit *)
                  pm := newPm;
                  RETURN Res.done;
                ELSE
                  (* ---------- fail during buffer allocation *)
                  EXIT;
                END;
              END;
            END;
            (* -------------------------- fail due to heap limitations *)
            WHILE pl > 0 DO
              DEC(pl);
              DrvMem.Free(base[pl].addr, length);
            END;
            DrvMem.Free(newPm, SIZE(ScrTyp.PixmapDescriptor));
            RETURN Res.noMemory;
          ELSE
            (* ---------------- fail: too large pixmap requested *)
            DrvMem.Free(newPm, SIZE(ScrTyp.PixmapDescriptor));
            RETURN Res.tooLargeRectangle;
          END;
        END;
      ELSE
        RETURN Res.noMemory;
      END;
    END New;

  PROCEDURE Dispose(VAR pm: PIXMAP);
    VAR
      prevPm: PIXMAP;
      pl: CARDINAL;
    BEGIN
      prevPm := PIXMAP(SrvObj.Prev(SrvObj.OBJECT(ScrVar.topPm), SrvObj.OBJECT(pm)));
      IF prevPm # NIL THEN
        WITH pm^ DO
          IF NOT screen THEN
            (* ---------------- deallocate the pixmap content and buffer *)
            DrvMem.Free(buffer.addr, width + 1);
            pl := PlanesPerPixel;
            WHILE pl > 0 DO
              DEC(pl);
              DrvMem.Free(base[pl].addr, length);
            END;
            (* ---------------------- exclude from the pixmap stack *)
            IF pm # ScrVar.topPm THEN
              prevPm^.next := next;
            ELSE
              ScrVar.topPm := next;
            END;
            (* --------------------- deallocate the descriptor *)
            DrvMem.Free(pm, SIZE(ScrTyp.PixmapDescriptor));
          END;
        END;
      END;
    END Dispose;

  PROCEDURE Copy(dst, src: PIXMAP);
    VAR
      pl: CARDINAL;
    BEGIN
      WITH dst^ DO
        IF screen THEN
          IF src^.screen THEN
            (* ---------- copy screen to screen: set registers *)
            ScrLow.SetWMode1();
            (* --------------------------------------- use byte moves! *)
            SrvLow.MoveInc(src^.base[0].addr, base[0].addr, length);
            (* ----------------------------------- reset registers *)
            ScrLow.ResetMode();
          ELSE
            (* --------------------- copy memory to screen *)
            FOR pl := 0 TO PlanesPerPixel - 1 DO
              (* -------------------------------- set registers *)
              ScrLow.SetMapMask(ShlMask[pl]);
              (* -------------------------------------- use byte moves! *)
              SrvLow.MoveInc(src^.base[pl].addr, base[0].addr, length);
            END;
            (* --------------------------- reset registers *)
            ScrLow.ResetMapMask();
          END;
        ELSE
          IF src^.screen THEN
            (* ------------------------ copy screen to memory *)
            FOR pl := 0 TO PlanesPerPixel - 1 DO
              (* ---------------------------------- set registers *)
              ScrLow.SetReadMap(SYSTEM.BYTE(pl));
              (* ------------------------------------- use byte moves! *)
              SrvLow.MoveInc(src^.base[0].addr, base[pl].addr, length);
            END;
          ELSE
            (* ------------------------------------- copy memory to memory *)
            FOR pl := 0 TO PlanesPerPixel - 1 DO
              SrvLow.FastMoveInc(src^.base[pl].addr, base[pl].addr, length);
            END;
          END;
        END;
      END;
    END Copy;

  (* ======================================= Graphics primitives *)

  PROCEDURE Clear(pm: PIXMAP; c: Scr.Pixel);
    VAR
      pl: CARDINAL;
    BEGIN
      WITH pm^ DO
        IF screen THEN
          (* --------------- clear on the screen using byte fills! *)
          ScrLow.SetWMode2();
          SrvLow.Fill(base[0].addr, length, BS(c));
          ScrLow.ResetMode();
        ELSE
          (* ---------------------- clear in the memory using word fills! *)
          FOR pl := 0 TO PlanesPerPixel - 1 DO
            IF pl IN BS(c) THEN
              SrvLow.FastFill(base[pl].addr, length, Full);
            ELSE
              SrvLow.FastFill(base[pl].addr, length, Null);
            END;
          END;
        END;
      END;
    END Clear;

  PROCEDURE Pt(pm: PIXMAP; p: Geo.Point; c: Scr.Pixel; func: Scr.Function);
    VAR
      pl, ofs1: CARDINAL;
      ptr: BP;
  (*$W+*)
      TM: BS;
  (*$W=*)
    BEGIN
      WITH pm^ DO
        (* ------------------------------------------ clipping *)
        IF view.clip AND NOT Geo.PtInRect(p, view.port) THEN
          RETURN;
        END;
        (* ---------------------------------------------- drawing ... *)
        IF screen THEN
          (* -------- ... on the screen: set the destination pointer *)
          ptr := base[0];
          INC(ptr.ofs,
            VAL(CARDINAL, p.iy) * width +
            VAL(CARDINAL, p.ix) DIV PixelsPerByte);
          (* --------------------------------------- set registers *)
          ScrLow.SetFuncWMode2BitMask
          (
            SYSTEM.BYTE(func),
            ShrMask[CARDINAL(p.ix) MOD PixelsPerByte]
          );
          (* ---------------------------------- set the byte *)
          TM := ptr.addr^;
          ptr.addr^ := BS(c);
          (* ------------------------- reset registers *)
          ScrLow.ResetFuncModeBitMask();
        ELSE
          (* ------ ... in the memory: set constant part of the offset *)
          ofs1 :=
            VAL(CARDINAL, p.iy - iy0) * width +
            VAL(CARDINAL, p.ix - ix0) DIV PixelsPerByte;
          (* ------------------------------------------- set the mask *)
          TM := BS(ShrMask[CARDINAL(p.ix - ix0) MOD PixelsPerByte]);
          (* ------------------------- loop through all the planes *)
          FOR pl := 0 TO PlanesPerPixel - 1 DO
            (* ------------------ set the destination pointer *)
            ptr := base[pl];
            INC(ptr.ofs, ofs1);
            (* ------------------- check function *)
            IF func = Scr.copy THEN
              (* -------------------- in replace mode *)
              IF pl IN BS(c) THEN
                ptr.addr^ := ptr.addr^ + TM;
              ELSE
                ptr.addr^ := ptr.addr^ - TM;
              END;
            ELSE
              (* ------------- in invert mode *)
              IF pl IN BS(c) THEN
                ptr.addr^ := ptr.addr^ / TM;
              END;
            END;
          END;
        END;
      END;
    END Pt;

  (*$O-,R-,Z-,I-*)

  PROCEDURE PtColor(pm: PIXMAP; p: Geo.Point): Scr.Pixel;
    VAR
      rx, ry, pl, ofs1: CARDINAL;
      ptr: BP;
      PM, CM: BS;
    BEGIN
      CM := Null;
      WITH pm^ DO
        IF screen THEN
          (* -------------------- set the pointer *)
          ptr := base[0];
          INC(ptr.ofs,
            VAL(CARDINAL, p.iy) * width +
            VAL(CARDINAL, p.ix) DIV PixelsPerByte);
          (* -------------------------------- set the pixel mask *)
          PM := BS(ShrMask[VAL(CARDINAL, p.ix) MOD PixelsPerByte]);
          (* ---------------------------------- read from the screen *)
          FOR pl := 0 TO PlanesPerPixel - 1 DO
            ScrLow.SetReadMap(SYSTEM.BYTE(pl));
            IF ptr.addr^ * PM # Null THEN
              INCL(CM, pl);
            END;
          END;
        ELSE
          (* ----------- compute relative coordinates *)
          rx := p.ix - ix0;
          ry := p.iy - iy0;
          (* -------------- set constant part of the offset *)
          ofs1 := ry * width + rx DIV PixelsPerByte;
          (* ---------------------------- set the pixel mask *)
          PM := BS(ShrMask[rx MOD PixelsPerByte]);
          (* ---------------------- read from the memory *)
          FOR pl := 0 TO PlanesPerPixel - 1 DO
            ptr := base[pl];
            INC(ptr.ofs, ofs1);
            IF ptr.addr^ * PM # Null THEN
              INCL(CM, pl);
            END;
          END;
        END;
      END;
      RETURN VAL(Scr.Pixel, ORD(CHAR(CM)));
    END PtColor;

  (*$O=,R=,Z=,I=*)

  PROCEDURE SolidRect(pm: PIXMAP; r: Geo.Rectangle; c: Scr.Pixel; func: Scr.Function);
    VAR
      lp: LineParams;
      pl, byte, lineCount, lineOfs, line, ofs1: CARDINAL;
      ptr: BP;
  (*$W+*)
      TM: BS;
  (*$W=*)
    BEGIN
      WITH pm^ DO
        (* -------------------------------------------------- clipping *)
        IF view.clip AND NOT Geo.RectXRectRes(r, view.port, r) THEN
          RETURN;
        END;
        (* ------------------------------------- set line parameters *)
        SetLineParams(r.ixMin - ix0, r.ixMax - ix0, lp);
        (* ----------------------------------- set the line count *)
        lineCount := r.iyMax - r.iyMin;
        (* ---------------------------------------- drawing ... *)
        IF screen THEN
          (* ------ ... on the screen: set the destination pointer *)
          ptr := base[0];
          INC(ptr.ofs,
            VAL(CARDINAL, r.iyMin) * width +
            VAL(CARDINAL, r.ixMin) DIV PixelsPerByte);
          (* ----------------------------------- set the line offset *)
          lineOfs := width - lp.byteCount;
          (* ---------------------------------------- set registers *)
          ScrLow.SetFuncWMode2BitMask(SYSTEM.BYTE(func), lp.LM);
          (* ------------------------------------------ main loop *)
          line := 0;
          LOOP
            (* ------------------------------ set the left byte *)
            TM := ptr.addr^;
            ptr.addr^ := BS(c);
            (* ------------------------------ set other bytes *)
            IF lp.byteCount > 0 THEN
              (* --------------------------- increment the offset *)
              INC(ptr.ofs);
              (* ----------------------------------- set inner bytes *)
              IF lp.byteCount > 1 THEN
                ScrLow.ResetBitMask();
                IF func = Scr.copy THEN
                  SrvLow.Fill(ptr.addr, lp.byteCount - 1, BS(c));
                ELSE
                  SrvLow.Refill(ptr.addr, lp.byteCount - 1, BS(c));
                END;
                INC(ptr.ofs, lp.byteCount - 1);
              END;
              (* ------------------------------ set the right byte *)
              ScrLow.SetBitMask(lp.RM);
              TM := ptr.addr^;
              ptr.addr^ := BS(c);
              (* ------------ reset the mask for the left byte *)
              ScrLow.SetBitMask(lp.LM);
            END;
            (* ---------------- check the stop condition *)
            IF line = lineCount THEN EXIT END;
            (* ------------------ increment the offset *)
            INC(ptr.ofs, lineOfs);
            (* --------- increment the line counter *)
            INC(line);
          END;
          (* ------------------- reset registers *)
          ScrLow.ResetFuncModeBitMask();
        ELSE
          (* ------ ... in the memory: set the constant part of the offset *)
          ofs1 :=
            VAL(CARDINAL, r.iyMin - iy0) * width +
            VAL(CARDINAL, r.ixMin - ix0) DIV PixelsPerByte;
          (* --------------------------------------------- main loop *)
          line := 0;
          LOOP
            (* -------------------- loop through all the planes *)
            FOR pl := 0 TO PlanesPerPixel - 1 DO
              (* -------------------- set destination pointer *)
              ptr := base[pl];
              INC(ptr.ofs, ofs1);
              (* -------------------- check function *)
              IF func = Scr.copy THEN
                (* ----------------------- in replace mode *)
                IF pl IN BS(c) THEN
                  (* --------------------- set the left byte *)
                  ptr.addr^ := ptr.addr^ + lp.LM;
                  (* --------------------------- set other bytes *)
                  IF lp.byteCount > 0 THEN
                    (* ----------------------- increment the offset *)
                    INC(ptr.ofs);
                    (* -------------------------------- set inner bytes *)
                    IF lp.byteCount > 1 THEN
                      SrvLow.FastFill(ptr.addr, lp.byteCount - 1, Full);
                      INC(ptr.ofs, lp.byteCount - 1);
                    END;
                    (* -------------------- set the right byte *)
                    ptr.addr^ := ptr.addr^ + lp.RM;
                  END;
                ELSE
                  (* --------------------- reset the left byte *)
                  ptr.addr^ := ptr.addr^ - lp.LM;
                  (* ------------------------- reset other bytes *)
                  IF lp.byteCount > 0 THEN
                    (* ------------------------ increment the offset *)
                    INC(ptr.ofs);
                    (* ------------------------------- reset inner bytes *)
                    IF lp.byteCount > 1 THEN
                      SrvLow.FastFill(ptr.addr, lp.byteCount - 1, Null);
                      INC(ptr.ofs, lp.byteCount - 1);
                    END;
                    (* ---------------- reset the right byte *)
                    ptr.addr^ := ptr.addr^ - lp.RM;
                  END;
                END;
              ELSE
                (* ------------------ in invert mode *)
                IF pl IN BS(c) THEN
                  (* ------------------------ invert the left byte *)
                  ptr.addr^ := ptr.addr^ / lp.LM;
                  (* -------------------- invert other bytes *)
                  IF lp.byteCount > 0 THEN
                    (* --------------------- increment the offset *)
                    INC(ptr.ofs);
                    (* ----------------------------- invert inner bytes *)
                    IF lp.byteCount > 1 THEN
                      SrvLow.FastFillXor(ptr.addr, lp.byteCount - 1, Full);
                      INC(ptr.ofs, lp.byteCount - 1);
                    END;
                    (* ---------------- invert the right byte *)
                    ptr.addr^ := ptr.addr^ / lp.RM;
                  END;
                END;
              END;
            END;
            (* ------------------ check the stop condition *)
            IF line = lineCount THEN EXIT END;
            (* ----------------- increment the offset *)
            INC(ofs1, width);
            (* --------- increment the line counter *)
            INC(line);
          END;
        END;
      END;
    END SolidRect;

  PROCEDURE StippledRect(pm: PIXMAP; r: Geo.Rectangle;
            stipple: SYSTEM.ADDRESS; w, h: CARDINAL; p: Geo.Point;
            c: Scr.Pixel; func: Scr.Function);

    VAR
      lp: LineParams;
      pl, byte, line, lineCount, lineOfs, ofs1: CARDINAL;
      stippleOfs0, stippleOfs1, stippleLineOfs: CARDINAL;
      srcOfs0, srcOfs1, delta, deltaC: CARDINAL;
      ptr, srcPtr, stipplePtr, buffIncPtr: BP;
      stippleLarge, fast: BOOLEAN;
  (*$W+*)
      TM: BS;
  (*$W=*)

    PROCEDURE ComputeSrc();
      VAR
        stipple: CARDINAL;
      BEGIN
        WITH pm^ DO
          IF stippleLarge THEN
            (* ------------ large stipple: check the shift count *)
            IF delta = 0 THEN
              (* --------------------------- copy without the shift *)
              SrvLow.FastMoveInc(stipplePtr.addr, buffer.addr, w);
            ELSE
              (* ------------------------------ copy with the right shift *)
              stipple := ORD(CHAR(stipplePtr.addr^));
              INC(stipplePtr.ofs, w - 1);
              SrvLow.MoveShDec(stipplePtr.addr, buffIncPtr.addr, w, delta);
              buffer.addr^ :=
                BS(SrvLow.Shr(stipple, delta)) +
                BS(SrvLow.Shl(ORD(CHAR(stipplePtr.addr^)), deltaC));
            END;
          ELSE
            (* --------- single-byte stipple: check the shift count *)
            IF delta = 0 THEN
              (* ----------------- copy single byte without the rotation *)
              buffer.addr^ := stipplePtr.addr^;
            ELSE
              (* ------------------ copy single byte with the rotation *)
              stipple := ORD(CHAR(stipplePtr.addr^));
              buffer.addr^ :=
                BS(SrvLow.Shr(stipple, delta)) +
                BS(SrvLow.Shl(stipple, deltaC));
            END;
          END;
        END;
      END ComputeSrc;

    PROCEDURE IncSrcOfs();
      BEGIN
        IF stippleLarge THEN
          INC(srcPtr.ofs);
          IF (srcPtr.ofs - srcOfs0) MOD w = 0 THEN
            DEC(srcPtr.ofs, w);
          END;
        END;
      END IncSrcOfs;

    BEGIN
      WITH pm^ DO
        (* --------------------------------------------------- clipping *)
        IF view.clip AND NOT Geo.RectXRectRes(r, view.port, r) THEN
          RETURN;
        END;
        (* ------------------------------------ set line parameters *)
        SetLineParams(r.ixMin - ix0, r.ixMax - ix0, lp);
        (* ----------------------------------- set the line count *)
        lineCount := r.iyMax - r.iyMin;
        (* -------------------------------- set stipple parameters *)
        stipplePtr.addr := stipple;
        stippleOfs0 := stipplePtr.ofs;
        stippleLineOfs := SrvMat.Mod(r.iyMin - p.iy, h);
        stippleLarge := w > 1;
        fast := vga AND NOT stippleLarge;
        stippleOfs1 := stippleOfs0;
        IF stippleLarge THEN
          INC(stippleOfs1, SrvMat.Mod(SrvMat.Div(r.ixMin - p.ix, PixelsPerByte), w));
        END;
        (* ----------------------------------- set shift parameters *)
        delta := SrvMat.Mod(p.ix, PixelsPerByte);
        deltaC := PixelsPerByte - delta;
        (* ------------------------------------- set source parameters *)
        IF (NOT screen) OR (NOT fast) THEN
          srcPtr.seg := buffer.seg;
          srcOfs0 := buffer.ofs;
          srcOfs1 := srcOfs0;
          IF stippleLarge THEN
            INC(srcOfs1, stippleOfs1 - stippleOfs0);
            IF delta # 0 THEN
              buffIncPtr := buffer;
              INC(buffIncPtr.ofs, w - 1);
            END;
          END;
        END;
        (* ------------------------------------- drawing ... *)
        IF screen THEN
          (* ------- ... on the screen: set the destination pointer *)
          ptr := base[0];
          INC(ptr.ofs,
            VAL(CARDINAL, r.iyMin) * width +
            VAL(CARDINAL, r.ixMin) DIV PixelsPerByte);
          (* ------------------------------------- set the line offset *)
          lineOfs := width - lp.byteCount;
          (* ------------------------------- is it possible to draw fast? *)
          IF fast THEN
            (* ------------------------------- yes: set registers *)
            ScrLow.VGASetFuncWMode3
            (
              SYSTEM.BYTE(func),
              SYSTEM.BYTE(delta),
              BS(c)
            );
            (* -------------------------------- main loop *)
            line := 0;
            LOOP
              (* ------------------------ set the stipple offset *)
              stipplePtr.ofs := stippleLineOfs * w + stippleOfs1;
              (* ---------------------------------- check the stipple *)
              IF stipplePtr.addr^ # Null THEN
                (* -------------- stipple is nonempty: set the left byte *)
                ScrLow.SetBitMask(lp.LM);
                TM := ptr.addr^;
                ptr.addr^ := stipplePtr.addr^;
                (* ------------------------------------- set other bytes *)
                IF lp.byteCount > 0 THEN
                  (* -------------------------- increment pointer offsets *)
                  INC(ptr.ofs);
                  (* ------------------------------------- set inner bytes *)
                  IF lp.byteCount > 1 THEN
                    ScrLow.ResetBitMask();
                    SrvLow.Refill(ptr.addr, lp.byteCount - 1, stipplePtr.addr^);
                    INC(ptr.ofs, lp.byteCount - 1);
                  END;
                  (* -------------------------------- set the right byte *)
                  ScrLow.SetBitMask(lp.RM);
                  TM := ptr.addr^;
                  ptr.addr^ := stipplePtr.addr^;
                END;
              ELSE
                (* -------- stipple is empty: increment the offset only *)
                INC(ptr.ofs, lp.byteCount);
              END;
              (* --------------------- check the stop condition *)
              IF line = lineCount THEN EXIT END;
              (* ----------- increment the destination offset *)
              INC(ptr.ofs, lineOfs);
              (* --------------- increment the stipple line *)
              stippleLineOfs := (stippleLineOfs + 1) MOD h;
              (* --------- increment the line counter *)
              INC(line);
            END;
            (* ------------------------- reset registers *)
            ScrLow.ResetFuncModeBitMask();
          ELSE
            (* --------------------------- no: set registers *)
            ScrLow.SetFuncWMode2(SYSTEM.BYTE(func));
            (* -------------------------------- main loop *)
            line := 0;
            LOOP
              (* -------------------- set the stipple offset *)
              stipplePtr.ofs := stippleLineOfs * w + stippleOfs0;
              (* ------------------------------------ check the stipple *)
              IF (stipplePtr.addr^ # Null) OR stippleLarge THEN
                (* --------- stipple is nonempty or large: compute the source *)
                ComputeSrc();
                (* ------------------------------ set the source offset *)
                srcPtr.ofs := srcOfs1;
                (* --------------------------------- set the left byte *)
                ScrLow.SetBitMask(srcPtr.addr^ * lp.LM);
                TM := ptr.addr^;
                ptr.addr^ := BS(c);
                (* ----------------------------- set other bytes *)
                IF lp.byteCount > 0 THEN
                  (* -------------------- increment pointer offsets *)
                  INC(ptr.ofs);
                  IncSrcOfs();
                  (* -------------------------------- set inner bytes *)
                  IF lp.byteCount > 1 THEN
                    (* ----------------------------- check the stipple *)
                    IF stippleLarge THEN
                      (* -------------------------- stipple is large *)
                      byte := 2;
                      LOOP
                        ScrLow.SetBitMask(srcPtr.addr^);
                        TM := ptr.addr^;
                        ptr.addr^ := BS(c);
                        INC(ptr.ofs);
                        IncSrcOfs();
                        IF byte = lp.byteCount THEN EXIT END;
                        INC(byte);
                      END;
                    ELSE
                      (* -------------------------------- stipple is small *)
                      ScrLow.SetBitMask(srcPtr.addr^);
                      SrvLow.Refill(ptr.addr, lp.byteCount - 1, BS(c));
                      INC(ptr.ofs, lp.byteCount - 1);
                    END;
                  END;
                  (* --------------------------------- set the right byte *)
                  ScrLow.SetBitMask(srcPtr.addr^ * lp.RM);
                  TM := ptr.addr^;
                  ptr.addr^ := BS(c);
                END;
              ELSE
                (* --- stipple is empty and not-large: increment the offset only *)
                INC(ptr.ofs, lp.byteCount);
              END;
              (* ------------------------ check the stop condition *)
              IF line = lineCount THEN EXIT END;
              (* ----------- increment the destination offset *)
              INC(ptr.ofs, lineOfs);
              (* --------------- increment the stipple line *)
              stippleLineOfs := (stippleLineOfs + 1) MOD h;
              (* --------- increment the line counter *)
              INC(line);
            END;
            (* ----------------------------------- reset registers *)
            ScrLow.ResetFuncModeBitMask();
          END;
        ELSE
          (* ---- ... in the memory: set the constant part of the offset *)
          ofs1 :=
            VAL(CARDINAL, r.iyMin - iy0) * width +
            VAL(CARDINAL, r.ixMin - ix0) DIV PixelsPerByte;
          (* ----------------------------------------- main loop *)
          line := 0;
          LOOP
            (* ------------------------- set the stipple offset *)
            stipplePtr.ofs := stippleLineOfs * w + stippleOfs0;
            (* ------------------------------------- check the stipple *)
            IF (stipplePtr.addr^ # Null) OR stippleLarge THEN
              (* --------- stipple is nonempty or large: compute the source *)
              ComputeSrc();
              (* ---------------- set the source offset *)
              srcPtr.ofs := srcOfs1;
              (* -------------------- loop through all the planes *)
              FOR pl := 0 TO PlanesPerPixel - 1 DO
                (* ------------------------ set the destination pointer *)
                ptr := base[pl];
                INC(ptr.ofs, ofs1);
                (* -------------------- check the function *)
                IF func = Scr.copy THEN
                  (* ------------------------------ in copy mode *)
                  IF pl IN BS(c) THEN
                    (* ----------------------------- set the left byte *)
                    ptr.addr^ := ptr.addr^ + srcPtr.addr^ * lp.LM;
                    (* --------------------------- set other bytes *)
                    IF lp.byteCount > 0 THEN
                      (* ----------------- increment offsets *)
                      INC(ptr.ofs);
                      IncSrcOfs();
                      (* -------------------------- set inner bytes *)
                      IF lp.byteCount > 1 THEN
                        (* ------------------------------ check the stipple *)
                        IF stippleLarge THEN
                          (* --------------------------------- stipple is large *)
                          byte := 2;
                          LOOP
                            ptr.addr^ := ptr.addr^ + srcPtr.addr^;
                            INC(ptr.ofs);
                            IncSrcOfs();
                            IF byte = lp.byteCount THEN EXIT END;
                            INC(byte);
                          END;
                        ELSE
                          (* ------------------------------------------ stipple is small *)
                          SrvLow.FastFillOr(ptr.addr, lp.byteCount - 1, srcPtr.addr^);
                          INC(ptr.ofs, lp.byteCount - 1);
                        END;
                      END;
                      (* ------------------------------ set the right byte *)
                      ptr.addr^ := ptr.addr^ + srcPtr.addr^ * lp.RM;
                    END;
                  ELSE
                    (* ----------------------------- reset the left byte *)
                    ptr.addr^ := ptr.addr^ - srcPtr.addr^ * lp.LM;
                    (* ------------------------------------ reset other bytes *)
                    IF lp.byteCount > 0 THEN
                      (* ----------------------- increment offsets *)
                      INC(ptr.ofs);
                      IncSrcOfs();
                      (* -------------------------- reset inner bytes *)
                      IF lp.byteCount > 1 THEN
                        (* ------------------------------- check the stipple *)
                        IF stippleLarge THEN
                          (* ---------------------------------- stipple is large *)
                          byte := 2;
                          LOOP
                            ptr.addr^ := ptr.addr^ - srcPtr.addr^;
                            INC(ptr.ofs);
                            IncSrcOfs();
                            IF byte = lp.byteCount THEN EXIT END;
                            INC(byte);
                          END;
                        ELSE
                          (* -------------------------------------------------- stipple is small *)
                          SrvLow.FastFillAnd(ptr.addr, lp.byteCount - 1, Full - srcPtr.addr^);
                          INC(ptr.ofs, lp.byteCount - 1);
                        END;
                      END;
                      (* --------------------------------- reset the right byte *)
                      ptr.addr^ := ptr.addr^ - srcPtr.addr^ * lp.RM;
                    END;
                  END;
                ELSE
                  (* ---------------------------- in invert mode *)
                  IF pl IN BS(c) THEN
                    (* --------------------------------- invert the left byte *)
                    ptr.addr^ := ptr.addr^ / (srcPtr.addr^ * lp.LM);
                    (* --------------------------- invert other bytes *)
                    IF lp.byteCount > 0 THEN
                      (* ---------------------- increment offsets *)
                      INC(ptr.ofs);
                      IncSrcOfs();
                      (* ------------------------ invert inner bytes *)
                      IF lp.byteCount > 1 THEN
                        (* ----------------------------- check the stipple *)
                        IF stippleLarge THEN
                          (* --------------------------------- stipple is large *)
                          byte := 2;
                          LOOP
                            ptr.addr^ := ptr.addr^ / srcPtr.addr^;
                            INC(ptr.ofs);
                            IncSrcOfs();
                            IF byte = lp.byteCount THEN EXIT END;
                            INC(byte);
                          END;
                        ELSE
                          (* ------------------------------------------- stipple is small *)
                          SrvLow.FastFillXor(ptr.addr, lp.byteCount - 1, srcPtr.addr^);
                          INC(ptr.ofs, lp.byteCount - 1);
                        END;
                      END;
                      (* ------------------------------- invert the right byte *)
                      ptr.addr^ := ptr.addr^ / (srcPtr.addr^ * lp.RM);
                    END;
                  END;
                END;
              END;
            END;
            (* -------------- check the stop condition *)
            IF line = lineCount THEN EXIT END;
            (* ----------- increment the destination offset *)
            INC(ofs1, width);
            (* ---------------- increment the stipple line *)
            stippleLineOfs := (stippleLineOfs + 1) MOD h;
            (* --------- increment the line counter *)
            INC(line);
          END;
        END;
      END;
    END StippledRect;

  PROCEDURE RectCopy(dst: PIXMAP; r: Geo.Rectangle;
            src: PIXMAP; p: Geo.Point; func: Scr.Function);
    VAR
      lp: LineParams;
      ox, oy, rxBegin, ryBegin, pxBegin, pyBegin: INTEGER;
      pl, line, lineCount, ofs1, srcOfs1, delta, deltaC: CARDINAL;
      forwardX, forwardY, extraLft, extraRht: BOOLEAN;
      ptr, srcPtr: BP;
      PM: BS;
  (*$W+*)
      TM: BS;
  (*$W=*)

    BEGIN
      WITH dst^ DO
        (* ----------------------------- clipping *)
        IF view.clip THEN
          ox := r.ixMin;
          oy := r.iyMin;
          IF Geo.RectXRectRes(r, view.port, r) THEN
            (* ------------------- shift of the source base point *)
            INC(p.ix, r.ixMin - ox);
            INC(p.iy, r.iyMin - oy);
          ELSE
            (* -------------- empty intersection *)
            RETURN;
          END;
        END;
        (* -------------------------- set line parameters *)
        SetLineParams(r.ixMin - ix0, r.ixMax - ix0, lp);
        (* ---------------------------- set the line count *)
        lineCount := r.iyMax - r.iyMin;
        (* --------------------------------- set direction flags *)
        forwardX := p.ix >= r.ixMin;
        forwardY := p.iy >= r.iyMin;
        (* ----------- set parameters according with X-direction flag *)
        IF forwardX THEN
          (* ---------------- ... beginning corners *)
          rxBegin := r.ixMin;
          pxBegin := p.ix;
          (* -------------------------------------- ... shift parameters *)
          delta := VAL(CARDINAL, p.ix - r.ixMin) MOD PixelsPerByte;
          IF delta # 0 THEN
            deltaC := PixelsPerByte - delta;
            extraLft := SrvMat.Mod(r.ixMin, PixelsPerByte) < deltaC;
            extraRht := SrvMat.Mod(r.ixMax, PixelsPerByte) >= deltaC;
          END;
        ELSE
          (* ----------------- ... beginning corners *)
          rxBegin := r.ixMax;
          pxBegin := p.ix + r.ixMax - r.ixMin;
          (* -------------------------------------- ... shift parameters *)
          delta := VAL(CARDINAL, r.ixMin - p.ix) MOD PixelsPerByte;
          IF delta # 0 THEN
            deltaC := PixelsPerByte - delta;
            extraRht := SrvMat.Mod(r.ixMax, PixelsPerByte) >= delta;
            extraLft := SrvMat.Mod(r.ixMin, PixelsPerByte) < delta;
          END;
        END;
        (* ----------- set parameters according with Y-direction flag *)
        IF forwardY THEN
          ryBegin := r.iyMin;
          pyBegin := p.iy;
        ELSE
          ryBegin := r.iyMax;
          pyBegin := p.iy + INTEGER(lineCount);
        END;
        (* -------------------------------------- copying *)
        IF
          screen AND
          src^.screen AND
          (lp.LM = Full) AND
          (NOT (lp.byteCount > 0) OR (lp.RM = Full)) AND
          (func = Scr.copy) AND
          (delta = 0)
        THEN
          (* ---------------- copy in read/write mode 1: set registers *)
          ScrLow.SetWMode1();
          (* -------------- set the destination pointer *)
          ptr := base[0];
          INC(ptr.ofs,
            VAL(CARDINAL, ryBegin) * width +
            VAL(CARDINAL, rxBegin) DIV PixelsPerByte);
          (* -------------------------- set the source pointer *)
          srcPtr := src^.base[0];
          INC(srcPtr.ofs,
            VAL(CARDINAL, pyBegin) * width +
            VAL(CARDINAL, pxBegin) DIV PixelsPerByte);
          (* ------------------------------------- main loop *)
          line := 0;
          LOOP
            (* -------------------- copy the line using byte moves *)
            IF forwardX THEN
              SrvLow.MoveInc(srcPtr.addr, ptr.addr, lp.byteCount + 1);
            ELSE
              SrvLow.MoveDec(srcPtr.addr, ptr.addr, lp.byteCount + 1);
            END;
            (* --------------------- check the stop condition *)
            IF line = lineCount THEN EXIT END;
            (* ---------------------------- change offsets *)
            IF forwardY THEN
              INC(ptr.ofs, width);
              INC(srcPtr.ofs, width);
            ELSE
              DEC(ptr.ofs, width);
              DEC(srcPtr.ofs, width);
            END;
            (* --------- increment the line counter *)
            INC(line);
          END;
          (* -------------------- reset registers *)
          ScrLow.ResetMode();
        ELSE
          (* ------------ copy in read/write mode 0: set registers *)
          IF screen THEN
            ScrLow.SetFunc(SYSTEM.BYTE(func));
          END;
          (* ------------------ set the destination offset *)
          ofs1 :=
            VAL(CARDINAL, ryBegin - iy0) * width +
            VAL(CARDINAL, rxBegin - ix0) DIV PixelsPerByte;
          (* --------------------------- set the source offset *)
          WITH src^ DO
            srcOfs1 :=
              VAL(CARDINAL, pyBegin - iy0) * width +
              VAL(CARDINAL, pxBegin - ix0) DIV PixelsPerByte;
          END;
          (* ------------------------------------------ main loop *)
          line := 0;
          LOOP
            (* -------------------- loop through all the planes *)
            FOR pl := 0 TO PlanesPerPixel - 1 DO
              (* ---------------- select the destination plane *)
              IF screen THEN
                ScrLow.SetMapMask(ShlMask[pl]);
              END;
              (* ---------------- set the destination pointer *)
              ptr := base[pl];
              INC(ptr.ofs, ofs1);
              (* ----------------------- select the source plane *)
              IF src^.screen THEN
                ScrLow.SetReadMap(SYSTEM.BYTE(pl));
              END;
              (* ------------------- set the source pointer *)
              srcPtr := src^.base[pl];
              INC(srcPtr.ofs, srcOfs1);
              (* ------------------------- check the X-direction flag *)
              IF forwardX THEN
                (* ------- in forward X-direction: check the shift count *)
                IF delta = 0 THEN
                  (* ------------ without shift: copy the left byte *)
                  IF screen THEN
                    ScrLow.SetBitMask(lp.LM);
                    PM := srcPtr.addr^;
                    TM := ptr.addr^;
                    ptr.addr^ := PM;
                  ELSIF func = Scr.copy THEN
                    ptr.addr^ := (ptr.addr^ - lp.LM) + srcPtr.addr^ * lp.LM;
                  ELSE
                    ptr.addr^ := ptr.addr^ / (srcPtr.addr^ * lp.LM);
                  END;
                  (* ----------------------------- copy other bytes *)
                  IF lp.byteCount > 0 THEN
                    (* ------------------------- increment offsets *)
                    INC(ptr.ofs);
                    INC(srcPtr.ofs);
                    (* ---------------- copy inner bytes using byte moves *)
                    IF lp.byteCount > 1 THEN
                      IF screen THEN
                        ScrLow.ResetBitMask();
                        IF func = Scr.copy THEN
                          SrvLow.MoveInc(srcPtr.addr, ptr.addr, lp.byteCount - 1);
                        ELSE
                          SrvLow.RemoveInc(srcPtr.addr, ptr.addr, lp.byteCount - 1);
                        END;
                      ELSIF func = Scr.copy THEN
                        SrvLow.MoveInc(srcPtr.addr, ptr.addr, lp.byteCount - 1);
                      ELSE
                        SrvLow.MoveXorInc(srcPtr.addr, ptr.addr, lp.byteCount - 1);
                      END;
                      INC(ptr.ofs, lp.byteCount - 1);
                      INC(srcPtr.ofs, lp.byteCount - 1);
                    END;
                    (* --------------------------- copy the right byte *)
                    IF screen THEN
                      ScrLow.SetBitMask(lp.RM);
                      PM := srcPtr.addr^;
                      TM := ptr.addr^;
                      ptr.addr^ := PM;
                    ELSIF func = Scr.copy THEN
                      ptr.addr^ := (ptr.addr^ - lp.RM) + srcPtr.addr^ * lp.RM;
                    ELSE
                      ptr.addr^ := ptr.addr^ / (srcPtr.addr^ * lp.RM);
                    END;
                  END;
                ELSE
                  (* --------------------- with shift: copy the left byte *)
                  IF extraLft THEN
                    PM := BS(SrvLow.Shl(ORD(CHAR(srcPtr.addr^)), delta));
                    IF (lp.byteCount > 0) OR extraRht THEN
                      INC(srcPtr.ofs);
                      PM := PM + BS(SrvLow.Shr(ORD(CHAR(srcPtr.addr^)), deltaC));
                    END;
                  ELSE
                    PM := BS(SrvLow.Shr(ORD(CHAR(srcPtr.addr^)), deltaC));
                  END;
                  IF screen THEN
                    ScrLow.SetBitMask(lp.LM);
                    TM := ptr.addr^;
                    ptr.addr^ := PM;
                  ELSIF func = Scr.copy THEN
                    ptr.addr^ := (ptr.addr^ - lp.LM) + PM * lp.LM;
                  ELSE
                    ptr.addr^ := ptr.addr^ / (PM * lp.LM);
                  END;
                  (* ---------------------------- copy other bytes *)
                  IF lp.byteCount > 0 THEN
                    (* ------------- increment the destination offset *)
                    INC(ptr.ofs);
                    (* ------------------------------- copy inner bytes *)
                    IF lp.byteCount > 1 THEN
                      IF screen THEN
                        ScrLow.ResetBitMask();
                        IF func = Scr.copy THEN
                          SrvLow.MoveShInc(srcPtr.addr, ptr.addr, lp.byteCount - 1, delta);
                        ELSE
                          SrvLow.RemoveShInc(srcPtr.addr, ptr.addr, lp.byteCount - 1, delta);
                        END;
                      ELSIF func = Scr.copy THEN
                        SrvLow.MoveShInc(srcPtr.addr, ptr.addr, lp.byteCount - 1, delta);
                      ELSE
                        SrvLow.MoveShXorInc(srcPtr.addr, ptr.addr, lp.byteCount - 1, delta);
                      END;
                      INC(ptr.ofs, lp.byteCount - 1);
                      INC(srcPtr.ofs, lp.byteCount - 1);
                    END;
                    (* ---------------------------------- copy the right byte *)
                    PM := BS(SrvLow.Shl(ORD(CHAR(srcPtr.addr^)), delta));
                    IF extraRht THEN
                      INC(srcPtr.ofs);
                      PM := PM + BS(SrvLow.Shr(ORD(CHAR(srcPtr.addr^)), deltaC));
                    END;
                    IF screen THEN
                      ScrLow.SetBitMask(lp.RM);
                      TM := ptr.addr^;
                      ptr.addr^ := PM;
                    ELSIF func = Scr.copy THEN
                      ptr.addr^ := (ptr.addr^ - lp.RM) + PM * lp.RM;
                    ELSE
                      ptr.addr^ := ptr.addr^ / (PM * lp.RM);
                    END;
                  END;
                END;
              ELSE
                (* ------- in backward X-direction: check the shift count *)
                IF delta = 0 THEN
                  (* ------------ without shift: copy the right byte *)
                  IF screen THEN
                    ScrLow.SetBitMask(lp.RM);
                    PM := srcPtr.addr^;
                    TM := ptr.addr^;
                    ptr.addr^ := PM;
                  ELSIF func = Scr.copy THEN
                    ptr.addr^ := (ptr.addr^ - lp.RM) + srcPtr.addr^ * lp.RM;
                  ELSE
                    ptr.addr^ := ptr.addr^ / (srcPtr.addr^ * lp.RM);
                  END;
                  (* ------------------------ copy other bytes *)
                  IF lp.byteCount > 0 THEN
                    (* ------------------------- decrement offsets *)
                    DEC(ptr.ofs);
                    DEC(srcPtr.ofs);
                    (* --------------- copy inner bytes using byte moves *)
                    IF lp.byteCount > 1 THEN
                      IF screen THEN
                        ScrLow.ResetBitMask();
                        IF func = Scr.copy THEN
                          SrvLow.MoveDec(srcPtr.addr, ptr.addr, lp.byteCount - 1);
                        ELSE
                          SrvLow.RemoveDec(srcPtr.addr, ptr.addr, lp.byteCount - 1);
                        END;
                      ELSIF func = Scr.copy THEN
                        SrvLow.MoveDec(srcPtr.addr, ptr.addr, lp.byteCount - 1);
                      ELSE
                        SrvLow.MoveXorDec(srcPtr.addr, ptr.addr, lp.byteCount - 1);
                      END;
                      DEC(ptr.ofs, lp.byteCount - 1);
                      DEC(srcPtr.ofs, lp.byteCount - 1);
                    END;
                    (* --------------------------- copy the right byte *)
                    IF screen THEN
                      ScrLow.SetBitMask(lp.LM);
                      PM := srcPtr.addr^;
                      TM := ptr.addr^;
                      ptr.addr^ := PM;
                    ELSIF func = Scr.copy THEN
                      ptr.addr^ := (ptr.addr^ - lp.LM) + srcPtr.addr^ * lp.LM;
                    ELSE
                      ptr.addr^ := ptr.addr^ / (srcPtr.addr^ * lp.LM);
                    END;
                  END;
                ELSE
                  (* ------------------- with shift: copy the right byte *)
                  IF extraRht THEN
                    PM := BS(SrvLow.Shr(ORD(CHAR(srcPtr.addr^)), delta));
                    IF (lp.byteCount > 0) OR extraLft THEN
                      DEC(srcPtr.ofs);
                      PM := PM + BS(SrvLow.Shl(ORD(CHAR(srcPtr.addr^)), deltaC));
                    END;
                  ELSE
                    PM := BS(SrvLow.Shl(ORD(CHAR(srcPtr.addr^)), deltaC));
                  END;
                  IF screen THEN
                    ScrLow.SetBitMask(lp.RM);
                    TM := ptr.addr^;
                    ptr.addr^ := PM;
                  ELSIF func = Scr.copy THEN
                    ptr.addr^ := (ptr.addr^ - lp.RM) + PM * lp.RM;
                  ELSE
                    ptr.addr^ := ptr.addr^ / (PM * lp.RM);
                  END;
                  (* ---------------------------- copy other bytes *)
                  IF lp.byteCount > 0 THEN
                    (* ------------- decrement the destination offset *)
                    DEC(ptr.ofs);
                    (* ------------------------------- copy inner bytes *)
                    IF lp.byteCount > 1 THEN
                      IF screen THEN
                        ScrLow.ResetBitMask();
                        IF func = Scr.copy THEN
                          SrvLow.MoveShDec(srcPtr.addr, ptr.addr, lp.byteCount - 1, delta);
                        ELSE
                          SrvLow.RemoveShDec(srcPtr.addr, ptr.addr, lp.byteCount - 1, delta);
                        END;
                      ELSIF func = Scr.copy THEN
                        SrvLow.MoveShDec(srcPtr.addr, ptr.addr, lp.byteCount - 1, delta);
                      ELSE
                        SrvLow.MoveShXorDec(srcPtr.addr, ptr.addr, lp.byteCount - 1, delta);
                      END;
                      DEC(ptr.ofs, lp.byteCount - 1);
                      DEC(srcPtr.ofs, lp.byteCount - 1);
                    END;
                    (* --------------------------------- copy the left byte *)
                    PM := BS(SrvLow.Shr(ORD(CHAR(srcPtr.addr^)), delta));
                    IF extraLft THEN
                      DEC(srcPtr.ofs);
                      PM := PM + BS(SrvLow.Shl(ORD(CHAR(srcPtr.addr^)), deltaC));
                    END;
                    IF screen THEN
                      ScrLow.SetBitMask(lp.LM);
                      TM := ptr.addr^;
                      ptr.addr^ := PM;
                    ELSIF func = Scr.copy THEN
                      ptr.addr^ := (ptr.addr^ - lp.LM) + PM * lp.LM;
                    ELSE
                      ptr.addr^ := ptr.addr^ / (PM * lp.LM);
                    END;
                  END;
                END;
              END;
            END;
            (* ----------------------- check the stop condition *)
            IF line = lineCount THEN EXIT END;
            (* ------------------------------ change offsets *)
            IF forwardY THEN
              INC(ofs1, width);
              INC(srcOfs1, src^.width);
            ELSE
              DEC(ofs1, width);
              DEC(srcOfs1, src^.width);
            END;
            (* ---------- increment the line counter *)
            INC(line);
          END;
          (* ----------------------------- reset registers *)
          IF screen THEN
            ScrLow.ResetFuncBitMapMask();
          END;
        END;
      END;
    END RectCopy;

  (* ============================= Partitions store/restore functions *)

  PROCEDURE PartsNmb(pm: PIXMAP): CARDINAL;
    BEGIN
      WITH pm^ DO
        RETURN PlanesPerPixel DIV number;
      END;
    END PartsNmb;

  PROCEDURE PartSize(pm: PIXMAP; n: CARDINAL): CARDINAL;
    BEGIN
      WITH pm^ DO
        RETURN length * number;
      END;
    END PartSize;

  PROCEDURE Store(pm: PIXMAP; n: CARDINAL; s: SYSTEM.ADDRESS);
    VAR
      pl: CARDINAL;
    BEGIN
      WITH pm^ DO
        (* ------------------------- check partition number *)
        n := n MOD (PlanesPerPixel DIV number);
        (* --------------------- store partition planes using byte moves *)
        FOR pl := n * number TO (n + 1) * number - 1 DO
          IF screen THEN ScrLow.SetReadMap(SYSTEM.BYTE(pl)) END;
          SrvLow.MoveInc(base[pl].addr, s, length);
          s := SrvSys.Add(s, length);
        END;
      END;
    END Store;

  PROCEDURE Restore(pm: PIXMAP; n: CARDINAL; s: SYSTEM.ADDRESS);
    VAR
      pl: CARDINAL;
    BEGIN
      WITH pm^ DO
        (* ----------------------------- check partition number *)
        n := n MOD (PlanesPerPixel DIV number);
        (* -------------------- restore partition planes using byte moves *)
        FOR pl := n * number TO (n + 1) * number - 1 DO
          IF screen THEN ScrLow.SetMapMask(ShlMask[pl]) END;
          SrvLow.MoveInc(s, base[pl].addr, length);
          s := SrvSys.Add(s, length);
        END;
        (* --------------------------- restore registers *)
        IF screen THEN ScrLow.ResetMapMask() END;
      END;
    END Restore;

  BEGIN
    vga := (ScrBas.adapter = ScrBas.vga);
  END ScrPl4.
