IMPLEMENTATION MODULE WinCur;
  (*
    Copyright (C) 1993-1995, Alex Iakovlev & Dimon Maslov
    **********************************************************************
    D.Maslov    Ver  1.0  Derived from module MarkCurs         April 91
    A.Iakovlev  Ver  2.0  Implementation of all the cursors    May 92
    A.Iakovlev  Ver  3.0  Use module Res                       August 92
    A.Iakovlev  Ver  4.0  Renaming                             June 93
    A.Iakovlev  Ver  5.0  Use type ScrTxOut.CharSet            January 94
    A.Iakovlev  Ver  6.0  New module ScrOut                    March 95
    **********************************************************************
  *)

  IMPORT Geo, Res, ScrXX, ScrOut, Win, WinOut;

  (* ============================================== Marker cursor *)

  TYPE MARKERCURSOR = POINTER TO ScrXX.MarkerCursorDescriptor;

  PROCEDURE NewMarker(dst: Win.WINDOW;
            a: ScrOut.Attrs; p: Geo.Point;
            VAR mc: MARKERCURSOR): Res.Status;
    BEGIN
      RETURN
        ScrXX.NewMarker
        (
          ScrXX.DESTINATION(dst),
          a, p, ScrXX.MarkProc(WinOut.Marker),
          ScrXX.MARKERCURSOR(mc)
        );
    END NewMarker;

  PROCEDURE MoveMarker(mc: MARKERCURSOR; p: Geo.Point);
    BEGIN
      ScrXX.MoveMarker(ScrXX.MARKERCURSOR(mc), p);
    END MoveMarker;

  PROCEDURE ChangeMarker(mc: MARKERCURSOR; a: ScrOut.Attrs);
    BEGIN
      ScrXX.ChangeMarker(ScrXX.MARKERCURSOR(mc), a);
    END ChangeMarker;

  PROCEDURE HideMarker(mc: MARKERCURSOR);
    BEGIN
      ScrXX.HideMarker(ScrXX.MARKERCURSOR(mc));
    END HideMarker;

  PROCEDURE ShowMarker(mc: MARKERCURSOR);
    BEGIN
      ScrXX.ShowMarker(ScrXX.MARKERCURSOR(mc));
    END ShowMarker;

  PROCEDURE Marker(mc: MARKERCURSOR; VAR a: ScrOut.Attrs; VAR p: Geo.Point);
    BEGIN
      ScrXX.Marker(ScrXX.MARKERCURSOR(mc), a, p);
    END Marker;

  PROCEDURE ExistMarker(mc: MARKERCURSOR): BOOLEAN;
    BEGIN
      RETURN ScrXX.ExistMarker(ScrXX.MARKERCURSOR(mc));
    END ExistMarker;

  PROCEDURE DisposeMarker(VAR mc: MARKERCURSOR);
    BEGIN
      ScrXX.DisposeMarker(ScrXX.MARKERCURSOR(mc));
    END DisposeMarker;

  (* ============================================= Rectangle Cursor *)

  TYPE RECTCURSOR = POINTER TO ScrXX.RectDescriptor;

  PROCEDURE NewRectangle(dst: Win.WINDOW;
            a: ScrOut.Attrs; r: Geo.Rectangle;
            VAR rc: RECTCURSOR): Res.Status;
    BEGIN
      RETURN
        ScrXX.NewRectangle
        (
          ScrXX.DESTINATION(dst),
          a, r, ScrXX.RectProc(WinOut.Rectangle),
          ScrXX.RECTCURSOR(rc)
        );
    END NewRectangle;

  PROCEDURE MoveRectangle(rc: RECTCURSOR; r: Geo.Rectangle);
    BEGIN
      ScrXX.MoveRectangle(ScrXX.RECTCURSOR(rc), r);
    END MoveRectangle;

  PROCEDURE ChangeRectangle(rc: RECTCURSOR; a: ScrOut.Attrs);
    BEGIN
      ScrXX.ChangeRectangle(ScrXX.RECTCURSOR(rc), a);
    END ChangeRectangle;

  PROCEDURE HideRectangle(rc: RECTCURSOR);
    BEGIN
      ScrXX.HideRectangle(ScrXX.RECTCURSOR(rc));
    END HideRectangle;

  PROCEDURE ShowRectangle(rc: RECTCURSOR);
    BEGIN
      ScrXX.ShowRectangle(ScrXX.RECTCURSOR(rc));
    END ShowRectangle;

  PROCEDURE Rectangle(rc: RECTCURSOR; VAR a: ScrOut.Attrs; VAR r: Geo.Rectangle);
    BEGIN
      ScrXX.Rectangle(ScrXX.RECTCURSOR(rc), a, r);
    END Rectangle;

  PROCEDURE ExistRectangle(rc: RECTCURSOR): BOOLEAN;
    BEGIN
      RETURN ScrXX.ExistRectangle(ScrXX.RECTCURSOR(rc));
    END ExistRectangle;

  PROCEDURE DisposeRectangle(VAR rc: RECTCURSOR);
    BEGIN
      ScrXX.DisposeRectangle(ScrXX.RECTCURSOR(rc));
    END DisposeRectangle;

  (* ============================================= Character cursor *)

  TYPE CHARCURSOR = POINTER TO ScrXX.CharCursorDescriptor;

  PROCEDURE NewChar(dst: Win.WINDOW;
            a: ScrOut.Attrs; p: Geo.Point; ch: CHAR;
            VAR cc: CHARCURSOR): Res.Status;
    BEGIN
      RETURN
        ScrXX.NewChar
        (
          ScrXX.DESTINATION(dst),
          a, p, ch, ScrXX.CharProc(WinOut.Char),
          ScrXX.CHARCURSOR(cc)
        );
    END NewChar;

  PROCEDURE MoveChar(cc: CHARCURSOR; p: Geo.Point);
    BEGIN
      ScrXX.MoveChar(ScrXX.CHARCURSOR(cc), p);
    END MoveChar;

  PROCEDURE ChangeChar(cc: CHARCURSOR; a: ScrOut.Attrs);
    BEGIN
      ScrXX.ChangeChar(ScrXX.CHARCURSOR(cc), a);
    END ChangeChar;

  PROCEDURE ReplaceChar(cc: CHARCURSOR; ch: CHAR);
    BEGIN
      ScrXX.ReplaceChar(ScrXX.CHARCURSOR(cc), ch);
    END ReplaceChar;

  PROCEDURE HideChar(cc: CHARCURSOR);
    BEGIN
      ScrXX.HideChar(ScrXX.CHARCURSOR(cc));
    END HideChar;

  PROCEDURE ShowChar(cc: CHARCURSOR);
    BEGIN
      ScrXX.ShowChar(ScrXX.CHARCURSOR(cc));
    END ShowChar;

  PROCEDURE Char(cc: CHARCURSOR; VAR a: ScrOut.Attrs; VAR p: Geo.Point; VAR ch: CHAR);
    BEGIN
      ScrXX.Char(ScrXX.CHARCURSOR(cc), a, p, ch);
    END Char;

  PROCEDURE ExistChar(cc: CHARCURSOR): BOOLEAN;
    BEGIN
      RETURN ScrXX.ExistChar(ScrXX.CHARCURSOR(cc));
    END ExistChar;

  PROCEDURE DisposeChar(VAR cc: CHARCURSOR);
    BEGIN
      ScrXX.DisposeChar(ScrXX.CHARCURSOR(cc));
    END DisposeChar;

  (* ================================================ String cursor *)

  TYPE STRINGCURSOR = POINTER TO ScrXX.StringCursorDescriptor;

  PROCEDURE NewString(dst: Win.WINDOW;
            a: ScrOut.Attrs; p: Geo.Point; s: ARRAY OF CHAR;
            VAR sc: STRINGCURSOR): Res.Status;
    BEGIN
      RETURN
        ScrXX.NewString
        (
          ScrXX.DESTINATION(dst),
          a, p, s, ScrXX.CharProc(WinOut.Char),
          ScrXX.STRINGCURSOR(sc)
        );
    END NewString;

  PROCEDURE MoveString(sc: STRINGCURSOR; p: Geo.Point);
    BEGIN
      ScrXX.MoveString(ScrXX.STRINGCURSOR(sc), p);
    END MoveString;

  PROCEDURE ChangeString(sc: STRINGCURSOR; a: ScrOut.Attrs);
    BEGIN
      ScrXX.ChangeString(ScrXX.STRINGCURSOR(sc), a);
    END ChangeString;

  PROCEDURE LengthString(sc: STRINGCURSOR): CARDINAL;
    BEGIN
      RETURN ScrXX.LengthString(ScrXX.STRINGCURSOR(sc));
    END LengthString;

  PROCEDURE ModSubString(sc: STRINGCURSOR; s: ARRAY OF CHAR; pos: CARDINAL; ins: BOOLEAN);
    BEGIN
      ScrXX.ModSubString(ScrXX.STRINGCURSOR(sc), s, pos, ins);
    END ModSubString;

  PROCEDURE DelSubString(sc: STRINGCURSOR; pos, len: CARDINAL);
    BEGIN
      ScrXX.DelSubString(ScrXX.STRINGCURSOR(sc), pos, len);
    END DelSubString;

  PROCEDURE SkipString(sc: STRINGCURSOR; pos: CARDINAL; set: ScrOut.CharSet; forward: BOOLEAN): CARDINAL;
    BEGIN
      RETURN ScrXX.SkipString(ScrXX.STRINGCURSOR(sc), pos, set, forward);
    END SkipString;

  PROCEDURE HideString(sc: STRINGCURSOR);
    BEGIN
      ScrXX.HideString(ScrXX.STRINGCURSOR(sc));
    END HideString;

  PROCEDURE ShowString(sc: STRINGCURSOR);
    BEGIN
      ScrXX.ShowString(ScrXX.STRINGCURSOR(sc));
    END ShowString;

  PROCEDURE String(sc: STRINGCURSOR; VAR a: ScrOut.Attrs; VAR p: Geo.Point; VAR s: ARRAY OF CHAR);
    BEGIN
      ScrXX.String(ScrXX.STRINGCURSOR(sc), a, p, s);
    END String;

  PROCEDURE ExistString(sc: STRINGCURSOR): BOOLEAN;
    BEGIN
      RETURN ScrXX.ExistString(ScrXX.STRINGCURSOR(sc));
    END ExistString;

  PROCEDURE DisposeString(VAR sc: STRINGCURSOR);
    BEGIN
      ScrXX.DisposeString(ScrXX.STRINGCURSOR(sc));
    END DisposeString;

  END WinCur.
